# Documentation at the __END__
package Data::CURL::CurlRequester;
#!/usr/bin/perl -w

use strict;
use warnings;
use Data::Dumper;
use feature qw(say);
use Tools::DependanciesViewer;
# COULD manage asynchronous events
# THE CRUD STEP : impossible to install under 12.04
# No user, no docs
# Use the WWW::Curl instead
#use Net::Curl::Simple;
use WWW::Curl::Simple;
use Net::Ping;
use LWP::UserAgent;
use version; our $VERSION = qv('0.2.1');

# This three one for Net::Async::HTTP version
# use Net::Async::HTTP;
# use IO::Async::Loop;
# use URI;


my $delimiter = "-";


# Prepares & return the request using an ids array
sub build_the_list_request{
    
    if ( $microlibrary::log_mode ){
    	print Dumper @_;
    }

	my ( $self, $nick, @ids_list ) = @_;
	# DEBUG
	#print "*********** Into CURLRequester ***********************************\n";
	#print Dumper(@ids_list);
	#print "NICK = $nick ***********************************\n";
	my $xml_data_to_process;

    if ( $microlibrary::log_mode ){
    	print "\n-- Preparing the request to Open Library ". $delimiter x50 . "\n";
	}
	# Parse the @ send as a parameter by the microlibrary.pl and sends the request
	my %rdf_fluxus = ();
	foreach ( @ids_list ){
		
		if ( $microlibrary::log_mode ){
			print "NICK = $nick ***********************************\n";
			print $_ . "\n";
		}
		
		my $request = "http://openlibrary.org/people/$nick/lists/$_.rdf";
		# DEBUG
		# print $request . "\n";

		# Should not call send_the_list_request()
		# see https://bugs.launchpad.net/bibliothek-project/+bug/856368
		$xml_data_to_process = &send_the_list_request( $request );
		# Put the data into the hash
		$rdf_fluxus{$_} = $xml_data_to_process;
	}
	# Returns a RDF fluxus hash : key = Open Library id / value = RDF flux
	return %rdf_fluxus;
}

# Prepares & return the request for only one cover
sub build_the_cover_request{

	my ( $self, $book_id ) = @_;
	# DEBUG
	say "*********** Into CURLRequester.build_the_cover_request ***********************************";
	# Remove the .rdf extension
	$book_id =~ s/.rdf//;
	#$string1 =~ s/apples/oranges/;
	say ($book_id);
	my $xml_data_to_process;

	say "-- Preparing the cover request to Open Library ". $delimiter x50;

	# You can add ?default=false to the URL to make it return 404 instead of blank image if cover doesn't exist
	# See the Covers API documentation for more details.
	# (don't use it)
	my $request = "http://covers.openlibrary.org/b/olid/$book_id-S.jpg";



	# DEBUG
	say $request;
	say "Cover request OK";
	return $request;
}

# Send the request to Open Library API
# Receive and returns the RDF fluxus to BUILD_THE_LIST_REQUEST() / To fix !!! \
sub send_the_list_request{
	my $request_to_send = shift;
	if ( $microlibrary::log_mode ){
		print "--- Get " . $request_to_send . " \n";
	}


	my $curl = WWW::Curl::Simple->new();
	# Provides a HTTP::Response object
	my $RDF_list_flux = $curl->get( $request_to_send )->content;

	# OLD WAY
	# my $curl = Net::Curl::Simple->new();
	# my $RDF_list_flux = $curl->get( $request_to_send )->{body};



	# IMPLEMENT TEST_REQUEST_STATUS() SUBROUTINE...

#    my $code = $curl->get("http://rueizorueorz.org/klm")->{code};
#    print $code;
#	die;


	# ASYNCHRONOUS TESTS
	# Asynchronous version with Net::Curl::Simple
	# Should work but don't, even if we compile libcurl 
	# with "--enable-ares" or "--enable-threaded-resolver"
	# See also : 
	# http://daniel.haxx.se/blog/2011/04/25/libcurls-name-resolving/
	# https://bugzilla.redhat.com/show_bug.cgi?id=514771	
	# Second asynchronous version with WWW::Curl::Multi (seems to be in development)
	# Try an asynchronous version with Net::Asynch::HTTP
	# http://search.cpan.org/~pevans/Net-Async-HTTP-0.11/lib/Net/Async/HTTP.pm
	# my $loop = IO::Async::Loop->new();
	# my $http = Net::Async::HTTP->new();

	# It would be very smart to test the status code returned !
	# http://openlibrary.org/dev/docs/restful_api#status_codes
	if ($RDF_list_flux){
		# DEBUG
		# print $rdf_flux;
		return $RDF_list_flux;
	}

	else{
		if ( $microlibrary::log_mode ){
			print "OUPS ! Your request gives shit...\n  Server down ?\n  Bad parameter ?\n";
		}
	}
}

# V0.1 : We load one time and we store this data :(
# We got the book URI from the book list into list RDF flux

# V0.2 : We really need asynchronous request here :)
# See Net::Curl::Simple::Async - perform Net::Curl requests asynchronously
# And tests into send_the_list_request()
sub send_the_book_request{
	my ( $self, $request_to_send) = @_;
	# DEBUG
	# print Dumper $request_to_send . "\n";


	my $curl= WWW::Curl::Simple->new();
	# Provides a HTTP::Response object
	my $RDF_book_flux = $curl->get( $request_to_send )->content;

	# OLD WAY
	# my $curl= Net::Curl::Simple->new();
	# my $RDF_book_flux = $curl->get( $request_to_send )->{body};

	# It would be very smart to test the status code returned here too !
    # http://openlibrary.org/dev/docs/restful_api#status_codes
	if ($RDF_book_flux){
		# DEBUG
		# print $RDF_book_flux;
		return $RDF_book_flux;
	}
	else{
		if ( $microlibrary::log_mode ){
			print "OUPS ! Your request gives shit...\n  Server down ?\n  Bad parameter ?\n";
		}
	}
}

sub send_the_cover_request{
	my ( $self, $request_to_send) = @_;
	# DEBUG
	say Dumper $request_to_send;
	say "OK" ;
	my $reachability = Net::Ping->new();
	my $book_cover;

	# Bug
	# Existing cover
	# Suppose that Open Library server is not down !!!

	say Dumper $reachability->ping($request_to_send);

	if ($reachability->ping('openlibrary.org')){

		say "OL Server reachable !!!";

		my $curl= WWW::Curl::Simple->new();
		# Provides a HTTP::Response object
		$book_cover = $curl->get( $request_to_send )->content;

		# OLD WAY
		# my $curl= Net::Curl::Simple->new();
		# $book_cover = $curl->get( $request_to_send )->{body};
		# DEBUG
		if ( $microlibrary::log_mode ){
			say Dumper $book_cover;
		}

		# It would be very smart to test the status code returned here too !
    	# http://openlibrary.org/dev/docs/restful_api#status_codes
		if ($book_cover){
			# DEBUG
			# print $RDF_book_flux;
			say "GOT COVER OK !!!";
			return $book_cover;
		}

		else{
			return $book_cover = 0;
		}

    } 
	# Not existing cover (404)
    elsif(!$reachability->ping($request_to_send)) {
    	$book_cover = 0;
    	say "Book cover = 0";
    	return $book_cover;
    }

	# DOWN !
	else {
		if ( $microlibrary::log_mode ){
			print "OUPS ! Your request gives shit...\n  Server down ?\n  Bad parameter ?\n";
		}
	}
}


sub test_request_status{
    my $ua = LWP::UserAgent->new();
    my $response = $ua->get('http://radiogrraphie.org/toto/titi.html');

    if ($response->is_success) {
         print $response->status_line;
         print $response->progress;
     }
     else {
         die $response->status_line;
     }
}

sub new {
    my $class = shift;
    my $self = {};
    bless ($self, $class);
    # DEBUG
	if ( $microlibrary::log_mode ){
		say "- New CurlRequester";
	}
    return $self;
}

if ( $microlibrary::sad_deps_mode ){
    my $env_viewer = Tools::DependanciesViewer->new();
    $env_viewer->print_used_modules($0, %INC);
}

1;


__END__


=pod

=head1 NAME                                                                                                                                                                                 

Data::CURL::CurlRequester - Build and send CURL requests to the Open Library (OL) RESTful API

=head1 DESCRIPTION

Contain tools to create and send a request to the Open Library RESTful API that
will return RDF fluxes or other things (books covers). Prepares data to send it to the View::WebDisplayer. 

The Data::CURL::CurlRequester wait for OL to return the flux : it's a quite long operation.
That's why you should use the $reload_mode flag in microlibrary.pl to tell it to 
call the Data::CURL::CurlRequester or not. Once you have run it one time, rdf fluxes
for lists and books are stored on your disk and it's not necessary to reload it 
each time you refresh your microlibrary page. 

See the documentation for the OL API here http://openlibrary.org/dev/docs/restful_api. 

=head1 CONFIGURATION AND ENVIRONMENT

There was a hardcoded parameter into C<< build_the_list_request() >> subroutine before.
It is now controlled by the mode.browser.simulation mode.
You just have to define the $nick scalar into microlibrary.pl when you need to reload data.

=head1 DEPENDENCIES

WWW::Curl::Simple

=head1 SUBROUTINES / METHODS

=over 4

=item B<< build_the_list_request() >>

Takes a list ids array as an argument. 
Build requests that are URI, contening an user name (hardcoded) and an id from the array.
This subroutine is bad-named, cause she calls the send_the_list_request() subroutine 
and returns an hash or rdf fluxes, where keys are the key are the OL list id and 
the value an RDF flux.

=item B<< send_the_list_request() >>

Send the request to Open Library API. Receive and returns the RDF fluxus to 
build_the_list_request(). This is a conception problem that we need to fix.

=item B<< send_the_book_request() >>

Receive the request from the Data::RDF::RDFTool and returns it too. 
Could test the OL server status code returned.

=back

=head1 ARGUMENTS

=over 4

=item B<< @ids_list >>

An array containing all the lists ids of an OL user.

=item B<< $request_to_send >>

The request to send is the value found in the list RDF flux. For each "seed" you've
got a value corresponding to the URI of a book

=back

=head1 BUGS AND LIMITATIONS

=over 4

=item B<< Subroutine role conception bug >>

https://bugs.launchpad.net/bibliothek-project/+bug/856368
send_the_list_request() receive and returns the RDF fluxus to 
build_the_list_request(). This is a conception problem that we need to fix.
Errors can be made cause build_the_list_request() don't do the job we can exept
according to it's name.

=item B<< Test returned status code >>

It would be very smart to test the status code returned.
https://blueprints.launchpad.net/bibliothek-project/+spec/daas2-ol-http-status-code
http://openlibrary.org/dev/docs/restful_api#status_codes

=item B<< Asynchronous CURL >>

We would like to make it asynchronous, so you don't have to wait for OL to return
 it's response to send another request.
 
=item B<< Multi-libraries >>

BULLSHITS !!! We are now on 0.02 version
At this 0.01 version, our Microlibrary is mono-library, so the name of the 
OL account is hard-coded into our module. It should be done otherwise (parsed from 
a properties file or taken into a database) and it will (multi-libraries). We will
 use a $OL_user scalar that will make possible to generate user-related requests. 

=back

=head1 AUTHOR

Self Shelf Swap - Simon Foutaiz - 2011-05-19

=head1 CURRENT MAINTAINER

    my ( $firstname = "smonff", $provider = "lavabit"); 
    print $firstname . "@" . $provider . "." .com;   

=head1 VERSION

This is version 0.02.1, well known as a Chinese Dime And A Shoestring

=head1 COPYRIGHT AND LICENSE  

Copyright 2011 Self Shelf Swap.

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl himself.

=cut
