# Documentation at the __END__
package Data::Cover::CoverTool;
#!/usr/bin/perl -w

use strict 'vars';
use warnings;
use IO qw(Handle File);
use Data::Dumper;
use feature qw(say);
use Data::CURL::CurlRequester;
use Tools::DependanciesViewer;
use Cwd;
use version; our $VERSION = qv('0.02');


my $cover = Data::CURL::CurlRequester->new();

# Takes the $simple_book_id as an argument ( URL with id without an extension)
sub get_cover{
	my ( $self, $cover_url, $book_id ) = @_;

	say "Into get_cover()";
	say $cover_url;
	say $book_id;

	my $cover_file = $cover->send_the_cover_request($cover_url);

	# No cover...
	if($cover_file eq 0){
		if ( $microlibrary::log_mode ){
			say "No cover for $book_id";
		}
	}

	# ... cover available
	else {
		# TODO
		# TEST 404 with ?default=false argument
		# it return 404 instead of blank image.
		
		if ( $microlibrary::log_mode ){
			say "Cover available for $book_id"
		}

		# GIF blank file for no-cover book
		if ($cover_file =~ /GIF89./ ){
			say "Hello blank file (no cover from Open Library)";
		} else {
			&store_cover($cover_file, $book_id);
		}
	}
	
}


# NOT USED ???
# REQUEST IS BUILD BY build_the_cover_request() into CurlRequester
sub get_cover_url{
	my ( $self, $OL_book_id ) = @_;
	my $cover_url = $cover->build_the_cover_request($OL_book_id);
	say $cover_url;
	return $cover_url;
}

########################################################
# !!!!!!!  if default void image, do not store !!!!!! ##
########################################################
sub store_cover{
	my  $cover_to_store = shift;
	my  $book_id = shift;
	# DEBUG
	# print Dumper($cover_to_store);

	chdir "./View/cover/" || die ("--- chdir error \n");
	
	my $current_path = cwd();

	if ( $microlibrary::log_mode ){
		say "Going to save $book_id cover to $current_path";
	}

	my $file_for_cover_storage = new IO::File "> $book_id.jpg";
	# TODO V0.03
	# We should test here if our file already exist and compare what is stored inside,
	# it's not necessary to re-write it if we already got the same thing...
	if ( defined $file_for_cover_storage ) {
		$file_for_cover_storage->print( $cover_to_store );

		if ( $microlibrary::log_mode ){
			print "--- $book_id cover serialized\n";
		}
		$file_for_cover_storage->close;
		if ( $microlibrary::log_mode ){
			print "--- $book_id cover closed\n";
		}
		chdir || die ('--- chdir error : you should fix the $ENV{HOME} directory inside Microlibrary Kernel \n');
	}
}

sub new {
	my $class = shift;
	my $self = {};
	bless ($self, $class);
	# DEBUG
	if ( $microlibrary::log_mode ){
		say "- New CoverTool";
	}
	return $self;
}


1;
__END__

=pod

=head1 NAME

Data::Cover::CoverTool

=head1 DESCRIPTION

Manipulates covers for book cover.openlibrary.org. It will use the 
send_the_cover_request() subroutine of the Data::CURL::CurlRequester for 
network request.


=head1 CONFIGURATION AND ENVIRONMENT

CoverTool will be used by the RDFTool to reload, store or use covers.
Is totally dependant of cover.openlibrary.org and it's API.
Will be used when the $reload_mode is set to '1'.

=head1 DEPENDENCIES

=over 4

=item B<< IO::Handle File >>

=item B<< Data::CURL::CurlRequester >>

=back

=head1 SUBROUTINES / METHODS

=over 4

=item B<< store_cover() >>
 Will store the cover to disk into the database after.

=item B<< get_cover >>
This is a trigger that will ask for the cover to the Open Library API, after 
build the URL for the request. It will ask fot the storage too.

=item B<< get_cover_url()) >>
Build the URL

=back

=head1 ARGUMENTS

=over 4

=item B<< URL >>

=back

=head1 BUGS AND LIMITATIONS

=over 4

=item Conception
The request is build into CurlRequester...

=back

=head1 AUTHOR

Self Shelf Swap - Smonff - 2012-03-06

=head1 CURRENT MAINTAINER

    my ( $firstname = "smonff", $provider = "lavabit"); 
    print $firstname . "@" . $provider . "." .com;   

=head1 VERSION

This is version 0.02.1, well known as a Chinese Dime And A Shoestring

=head1 COPYRIGHT AND LICENSE  

Copyright 2011 Self Shelf Swap.

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl himself.

=cut
