# Documentation at the __END__
package Data::RDF::RDFTool;
#!/usr/bin/perl -w

use strict 'vars';
use warnings;
use IO qw(Handle File);
use Data::Dumper;
use IO::File;
use XML::Simple;
use Data::CURL::CurlRequester;
use File::Basename;
use View::WebDisplayer;
use Encode::Encoder;
use Tools::DependanciesViewer;
use version; our $VERSION = qv('0.2.1');
use Data::Cover::CoverTool;
use feature qw(say);
use Perl6::Slurp;
use Text::Buffer;

my $book_element;
my $element;
my $list_flux_transmitted;
# my $book_flux_transmitted;
my $tmp_list_element;
my $tmp_book_element;
my $tmp_list_description;
my $list_description;
my $tmp_list_name;
my $list_name;
my $tmp_list_url;
my $list_url;
# Used to store the extracted list elements
my %list_elements;
my $list_selector;
my @book_element_selector;
my %books_data;
my $tmp_book_description;
my @tmp_book_descriptions;
my %tmp_data_for_display = ();
my $list_name_only;




sub new {
	my $class = shift;
	my $self = {};
	bless ($self, $class);
	# DEBUG
	if ( $microlibrary::log_mode ){
		say "- New RDFTool";
	}
	return $self;
}

# View stuff (should move soon) #################################################
# SAD (Simple And Dirty) way to do it :(
# We should do better = not mix view and model...
my $view = View::WebDisplayer->new();
#################################################################################

# Serialize RDF flux retrieved by the Curl Requester from Open Library
# Books lists or books are possible.
sub store_rdf_flux{
	my (
	    $self, 
	    $flux_to_store,
	    $flux_id, # Book or list id
	    $list_or_book_selector # used to tell wich kind of RDF flux we manage
		) = @_;
	# DEBUG
	# print Dumper($flux_to_store, $flux_id);

	if ( $list_or_book_selector eq "list" ){
		$flux_id = $flux_id . ".rdf";

		if ( $microlibrary::log_mode ){
			print "--- Flux $flux_id is a list\n";
		}
		chdir "../database/list/" || die ("--- chdir error \n");
	}

	elsif ( $list_or_book_selector eq "book" ){
		if ( $microlibrary::log_mode ){
			print "--- Flux $flux_id is a book\n";
		}
		chdir "../database/book/" || die ("--- chdir error \n");
	}


	my $file_for_list_storage = new IO::File "> $flux_id";
	# TODO V0.03
	# We should test here if our file already exist and compare what is stored inside,
	# it's not necessary to re-write it if we already got the same thing...
	if ( defined $file_for_list_storage ) {
		$file_for_list_storage->print( $flux_to_store );

		if ( $microlibrary::log_mode ){
			print "--- $flux_id serialized\n";
		}
		$file_for_list_storage->close;
		if ( $microlibrary::log_mode ){
			print "--- $flux_id closed\n";
		}
		chdir || die ('--- chdir error : you should fix the $ENV{HOME} directory inside Microlibrary Kernel \n');
	}
}

# This is the lists and books parser
# Parse RDFFlux from stored files using the XML::Simple parser
# http://search.cpan.org/~mthurn/RDF-Simple-0.415/lib/RDF/Simple/Parser.pm
sub parse_RDF_file{

	my ( $self, $OL_list_id ) = @_;
	# DEBUG
	# print $OL_list_id . "\n";

	# Build the name of the file
	my $rdf_file_name = $OL_list_id . ".rdf";
	if ( $microlibrary::log_mode ){
		print "\n--- Get list " . $rdf_file_name . " for parsing \n";
	}

	chdir "../database/list/" || die ("--- chdir error \n");
	my $xml_flux_to_parse = XML::Simple->new();

	# / ! \This is a hash of array handling our file...
	my $flux_to_parse_ref = $xml_flux_to_parse->XMLin("$rdf_file_name");
	# print Dumper $flux_to_parse_ref;

	# Get back to the $ENV{HOME} directory
	chdir || die ("--- chdir error \n");

	# Populate elements of an hash containing all informations about a RDF list
	# v0.01 : return all data into a hash containing the list name, it's description, etc.
	#         using a subroutine get_list_element("element") where element is the XML node.
	#         Value is accessed by passing the element name as an argument
	# TODO v0.03 : MOOSE !
	if ( $microlibrary::log_mode ){
	   print "--- Get descriptors for this list\n";
	}

	# List URI : something like http://openlibrary.org + <ol:key>/people/sandoth/lists/OL6097L</ol:key>
    $list_selector = "key";
    $list_url = &get_list_element( $flux_to_parse_ref, "$list_selector" );

    if ( $list_url eq -1 ){
        if ( $microlibrary::log_mode ){
        	print "---- No $list_selector for $rdf_file_name \n";
        }
    }

    else{
        $list_url = "http://openlibrary.org" . $list_url;   # Build the entire URI
        $list_elements{ "$list_selector" } = $list_url;     # Store entire URI in our catchall hash

        if ( $microlibrary::log_mode ){
        	print "---- List $list_selector added to %list_elements \n";
            print "----- $list_selector = " . $list_elements{ $list_selector } . "\n";
        }
    }

    # List name
    $list_selector = "name";
    $list_name = &get_list_element( $flux_to_parse_ref, "$list_selector");
    #say "LIST NAME = " . $list_name;


    # Get number of shelves
    my $shelves_size = @microlibrary::shelves;

    # Put list titles into the global array
    my $array_hut = $shelves_size + 1;
    if (defined $list_name){
		$microlibrary::shelves[$array_hut] = $list_name;    	
    }



    if ( $list_name eq -1 ){
        if ( $microlibrary::log_mode ){
        	print "---- No $list_selector for $rdf_file_name \n";
        }
    }

    else{
        $list_elements{ "$list_selector" } = $list_name;
        if ( $microlibrary::log_mode ){
        	print "---- List $list_selector added to %list_elements \n";
        	print "----- $list_selector = " . $list_elements{ $list_selector } . "\n";
        }
    }

    # List description
    $list_selector = "description";
    $list_description = &get_list_element( $flux_to_parse_ref, "$list_selector");

    if ( $list_description eq -1 ){
        if ( $microlibrary::log_mode ){
            print "---- No $list_selector for $rdf_file_name \n";
        }
    }


    # View stuff (should move soon for v0.03) #############################################
    if ( $microlibrary::html_mode ){
	    $view->newline();
        $view->open_tag( "div", "class", "list-info" );
	    #$view->open_tag( "h2", "class", "list-title"  );
        #$microlibrary::html_buffer->append( $list_name );
	    # $view->print_into_tags( $list_name, "a", "href", $list_url );
		#$view->close_tag();
		#$view->newline();
		if( $list_description ne -1 ){
            $view->print_into_tags("$list_description", "p", "class", "list-description");
		}
		$view->close_tag();



		#$view->open_tag( "div", "class", "books-list"  );
	}############################################################################






	else{
		$list_elements{ "$list_selector" } = $list_description;
		# TODO !!!
		# $microlibrary::context{main_list_description} = $list_description;
		if ( $microlibrary::log_mode ){
			print "---- List $list_selector added to %list_elements \n";
		    print "----- $list_selector = " . $list_elements{ $list_selector };
		}
	}

	# This is the book part.
	# Should maybe be in another subroutine, cause parse_ref_file() calls lists
	# books stuffs, that's not a good conception pattern but we werer in a hurry...
	if ( $microlibrary::log_mode ){
		print "\n--- Build the requests\n";
	}

	my $books_quantity_calculator = &get_books_number_inside_list($flux_to_parse_ref);
	# How to calculate the number of items in an hash of arrays
	# http://www.linuxquestions.org/questions/programming-9/how-do-i-get-number-of-elements-in-a-hash-of-arrays-689620/
	# Get the length of the dereferenced array reference
	# That's a little bit tricky, you can't invent it
	my $books_quantity = $#$books_quantity_calculator + 1;
	$microlibrary::context{'items_nbr'} = $books_quantity;

	if ( $microlibrary::log_mode ){
	  print Dumper %microlibrary::context;
	  print Dumper $flux_to_parse_ref;
	}

	if ( $microlibrary::log_mode ){
		print "---- Got " . $books_quantity . " items in $OL_list_id\n";
	}

	my $book_request_url = "";
	my $book_id = "";
	my $books_data_from_ol = Data::CURL::CurlRequester->new();
	my $books_flux_returned; 
	my $processed_book_template;
	my $book_text_buffer = Text::Buffer->new();

	# Will treat all the books of a list
	# Maybe the bug for lists with only one book is here
	for ( my $counter = 0; $counter < $books_quantity;  ) {

	  # Get book data
	  $book_request_url = &get_book_request_url( $flux_to_parse_ref, $counter );
	  $book_id = &get_book_id( $book_request_url ) ;
	  # DEBUG
	  # print " !!!!!!!!!!!!!!!!!!!!!!! BOOK $book_id \n";

	  $counter++;

	  # Remove the .rdf extension
	  my $simple_book_id = $book_id;
	  $simple_book_id =~ s/.rdf//;
	  my $cover_path;

	  if( $microlibrary::reload_mode ){
          # Will send books URI values to the books CURL requester and then...
	    my $books_flux_returned = $books_data_from_ol->send_the_book_request( $book_request_url );
	    # Caches the RDF flux of the book into a file !
	    &store_rdf_flux( $self, $books_flux_returned, $book_id, "book"  );

	    # Get the cover and store it
	    my $cover = Data::Cover::CoverTool->new();
	    my $cover_url = $cover->get_cover_url($simple_book_id);
	    say $cover_url;
	    my $cover_data = $cover->get_cover($cover_url, $simple_book_id);

	  }

	  # Render books data with get_book_element() :
	  # 1 -> We got the lists
	  # 2 -> Make a RDF file file handle
	  # 3 -> Render into get_book_element

        # Go searching the book
        my $rdf_book_file_name = $book_id;
        #print "\n--- Get book " . $rdf_book_file_name . " for parsing \n";
	    chdir "../database/book/" || die ("--- chdir error \n");
	    my $xml_book_flux_to_parse = XML::Simple->new();

	    # This is simply a hash of array...
	    my $book_flux_to_parse_ref = $xml_book_flux_to_parse->XMLin("$rdf_book_file_name");
        # print Dumper $book_flux_to_parse_ref;
	    chdir || die ("--- chdir error \n");

	    # Get diferent interesting elements
	    my $book_title = &get_book_element( "title", $book_flux_to_parse_ref);
	    my $book_editor = &get_book_element( "publisher", $book_flux_to_parse_ref);
        my $book_url = &get_book_element( "url", $book_flux_to_parse_ref);
        my $book_issue = &get_book_element( "issued", $book_flux_to_parse_ref);
        # Would be good to work on an author REST requester and fix the 'year remplacing author' bug
        my $book_author = &get_book_element( "author", $book_flux_to_parse_ref);

        # Get the cover
        #say $simple_book_id;
		my $cover_name = $simple_book_id . ".jpg";
		#say $cover_name;

		my $book_vars = {
			book 	=>{
				id 				=> $simple_book_id,
				title			=> $book_title,
				author 			=> $book_author,
				editor			=> $book_editor,
				issue			=> $book_issue,
				url				=> $book_url,
			},
		};


		#my $current_dir = `ls -al`;
		#print $current_dir and die;
		chdir "./View/cover/" || die ("--- chdir error \n");

		if( -e $cover_name){
			$cover_path = "./View/cover/$cover_name";
			#say $cover_path and die;
		} #else {
			#$cover_path = "./View/cover/generic_cover.svg";
		#}


		# say Dumper $book_vars;


		# Add the book cover path to the hash
		$$book_vars{book}{cover_path} = $cover_path;

		chdir || die ('--- chdir error : you should fix the $ENV{HOME} directory inside Microlibrary Kernel \n');

  
        if ( $microlibrary::html_mode ){

        	################################
        	#
        	# 		DA BOOK TEMPLATE 
        	#
        	#		using 'the Stash' could be a great thing
        	#		See Perl TT official documentation
        	#
        	################################

        	my $library_book_template = Template->new($microlibrary::templates_conf);

        	my $book_template_file = 'book.tmpl';

   	        my $tmp_book_template_buffer = '';

   	        # say Dumper $book_vars;

   	        # say Dumper ($library_book_template, $book_template_file, $book_vars) and die;

        	$library_book_template->process($book_template_file, $book_vars, 'tmp_book_template_buffer')
        		|| die $library_book_template->error("Can't process microlibrary book template :(");

        	if ( $microlibrary::log_mode ){
        		say "Template processed";
        	}


        	my $book_file_handle = FileHandle->new;
	        if ($book_file_handle->open("< $microlibrary::tmp_dir/tmp_book_template_buffer")) {
	          if ( $microlibrary::log_mode ){
	            say "File open";
	          }
	          my $slurped_file = slurp($book_file_handle);
	          
              	# add the book to a Text::Buffer  (concatenate at the end of the file)
	        	$book_text_buffer->append($slurped_file);
	        }
	        else{
	          die("Cannot ssslurp processed template file");
	        }
			

			$microlibrary::context{shelf}{name} = $list_name;
			$microlibrary::context{shelf}{name} = $list_description;

        }
	}

	# Append only when the book loop is end
	$view->append($book_text_buffer->dumpAsString());

	#$view->close_tag();
}


# Get an element from the RDF list flux
sub get_list_element{
	$list_flux_transmitted = shift;
	$element = shift;
	if ( $list_flux_transmitted->{ "ol:$element" } ){
		$tmp_list_element = $list_flux_transmitted->{ "ol:$element" };
	}

	else {
		$tmp_list_element = -1;
	}

	return $tmp_list_element;
}

# Get an element from the RDF book flux
sub get_book_element{

    $element = shift;
    my $book_flux_transmitted = shift;
    #my $book_id = shift;
    #print Dumper($book_flux_transmitted) . "\n";

    # See seeds 
    # Get list elements
    # MUST dereference $$ -> works on line 
    if ( $element eq "title" || 
            $element eq "publisher" || 
            $element eq "issued" || 
            $element eq "extent" ||
            $element eq "language"  ) {

        if ( $microlibrary::log_mode ){
        	print Dumper $book_flux_transmitted;
            print $element . "\n";
        }

        if ( defined $$book_flux_transmitted{ 'rdf:Description' }->{ "dcterms:$element" } ){

            # https://bugs.launchpad.net/bibliothek-project/+bug/858280
            if ( $element eq "publisher" ){
	            # Sometimes, their is many publishers, so we got an array
	            # my $ref_type = ref $$book_flux_transmitted{ 'rdf:Description' }->{ "dcterms:$element" };
	            # print $ref_type;
	            # print $element;

	            # BUG !!! Can't retrieve array elements
	            if ( ref($$book_flux_transmitted{'rdf:Description' }->{ "dcterms:$element" }) eq "ARRAY" ) {
	                # @tmp_book_descriptions = $$book_flux_transmitted{ 'rdf:Description' }->{ "dcterms:$element" };
	                # print Dumper @tmp_book_descriptions;
	                # print @tmp_book_descriptions;
					# foreach $k ( @tmp_book_descriptions ){
					# print $k . "\n ___________________________________";
	                # $tmp_book_description = $tmp_book_description[] . " - " . $_ ;  
	                #print $_ . "\n";
	                # print $tmp_book_description;
	                # }
	                $tmp_book_description = "--";

	            }

	            else {
	               $tmp_book_description = $$book_flux_transmitted{ 'rdf:Description' }->{ "dcterms:$element" };
	            }
            }

            else{
                $tmp_book_description = $$book_flux_transmitted{ 'rdf:Description' }->{ "dcterms:$element" };
            }
        }

        # Ne marche pas
        else {
        	# print "No element value for this book";
        	$tmp_book_description = "--";
        }

        if ( $microlibrary::log_mode ){
            print "///////////////// tmp_book_description  = $tmp_book_description /////////////////\n";
        }
}

    elsif( $element eq "author"){
       # https://bugs.launchpad.net/bibliothek-project/+bug/858280
       # V0.2
       # If the book have many authors, the authors are stored into an array,
       # not a hash ( e.g. : OL8567763M.rdf", "OL7361700M", etc.)
       # We also encounter probelms with mono-author book.
       # Waiting for an request send to the Open Library team.
       # Added a patch

       # ref() function allows to know which type of var we use
       # http://www.perlmonks.org/?node=ref
	   if ( ref($$book_flux_transmitted{'rdf:Description' }->{ 'bibo:authorList' }->{ 'rdf:Description' }) eq "HASH" ){

	    	if ( defined $$book_flux_transmitted{ 'rdf:Description' }->{ 'bibo:authorList' }->{ 'rdf:Description' } ){
		    	$tmp_book_description = $$book_flux_transmitted{ 
		    		'rdf:Description' }->{ 'bibo:authorList' }->{ 'rdf:Description' }->{ 'rdf:value' };
		    	# DEBUG
		    	# print "=== AUTHOR : " . $$book_flux_transmitted{ 'rdf:Description' }->{ 'bibo:authorList' }->{ 'rdf:Description' }->{ 'rdf:value' } . "\n";
	    	}

	    	else {
	    		$tmp_book_description = "--";
	    	}
	    }

	   elsif ( ref($$book_flux_transmitted{'rdf:Description' }->{ 'bibo:authorList' }->{ 'rdf:Description' }) eq "ARRAY" ){
	   	   # print "BUG : arrayed author !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";
	       $tmp_book_description = "--";
	   }

	   # Temporary patch for bug #858280
	   # If nothing can be done, we can get the authors information into the 'work' page, not the 'book' one
	   elsif( defined $$book_flux_transmitted{'rdf:Description' }->{ 'bibo:authorList' }->{ 'rdf:parseType' } ){
	   	   $tmp_book_description = "many authors";
	   }

	   # "Unknow author", cause we don't have any value under "bibo:authorList" 
	   else{
	   	   $tmp_book_description = "--";
	   }
    }

    elsif( $element eq "edition" ){
    	$tmp_book_description = $$book_flux_transmitted{ 'rdf:Description' }->{ "bibo:$element" };
    }

    elsif( $element eq "url" ){
    	$tmp_book_description = $$book_flux_transmitted{ 'rdf:Description' }->{ "rdf:about" };
    }
    # print $element;
    # print Dumper %books_data;

    # Debug for author troubles
    # if( $element eq "author" ){
    #    print '$tmp_book_element before return__________________' . $tmp_book_description . "\n";
    #}

    # Try to avoid "wide character" warning
    my $encoded_tmb_book_des = Encode::encode_utf8($tmp_book_description);
    return $encoded_tmb_book_des;}

sub get_books_number_inside_list{
	my $flux_to_parse_ref_transmission = shift;

    # How to calculate the number of items in an hash of arrays  
    # http://www.linuxquestions.org/questions/programming-9/how-do-i-get-number-of-elements-in-a-hash-of-arrays-689620/
    my $books_number = $flux_to_parse_ref_transmission->{ "ol:seeds" };
	return $books_number;
}

# The counter is passed as a parameter that's incremented on each loop.
sub get_book_request_url{
    my $flux_to_parse_ref_transmission = shift;
    my $tmp_counter = shift;
    my $tmp_book_request_url;
    # DEBUG
    # print $counter . "\n";
    # print $flux_to_parse_ref->{ "ol:seeds" }[ $counter ] . "\n";
    # We directly go for searching books URI into the RDF book lists
    $tmp_book_request_url = $flux_to_parse_ref_transmission->{ "ol:seeds" }[ $tmp_counter ];
    if ( $microlibrary::log_mode ){
        print "----- Book URL : " . $tmp_book_request_url . "\n";
    }

    return $tmp_book_request_url;
}

sub get_book_id{

	my $tmp_book_request_url = shift;
    # We don't know it, we have to get it from the URI
	my $tmp_book_id;

        # Retrieve the book id from an URL :
        # We need a string starting with "OL" and ending by ".rdf"
        $tmp_book_id =  basename( $tmp_book_request_url );
        if ( $microlibrary::log_mode ){
        	print "----- Book id : " . $tmp_book_id . "\n";
        }

	return $tmp_book_id;
}

if ( $microlibrary::sad_deps_mode ){
    my $env_viewer = Tools::DependanciesViewer->new();
    $env_viewer->print_used_modules($0, %INC);
}

1;
__END__

=pod

=head1 NAME

Data::RDF::RDFTool

=head1 DESCRIPTION

Manipulates RDF fluxus or files. Le coeur de la meule.

This is the swiss army knife for our RDF flux and files.
We can easyly parse RDF flux, store it into files, and get many elements into 
this flux. We can also retrieve data from the files and use it for displaying the 
items of our library.

=head1 CONFIGURATION AND ENVIRONMENT

Plays with directory and $ENV{HOME}, so this ENV var should be fixed into the 
Microlibrary Kernel.


=head1 DEPENDENCIES

=over 4

=item B<< IO::Handle File >>

=item B<< IO::File >>

=item B<< XML::Simple >>

=item B<< File::Basename >>

=item B<< Encode::Encoder >>

=item B<< Data::CURL::CurlRequester >>

=item B<< View::WebDisplayer >>

=back

=head1 SUBROUTINES / METHODS

=over 4

=item B<< store_rdf_flux() >>

Serialize RDF flux retrieved by the Curl Requester from Open Library.
Books lists or books are possible but this must be specified as an argument.
Should manage undesirable behaviours !!!

=item B<< parse_RDF_file(OL_list_id) >>

Takes an OL list of books file id in argument from the Microlibrary Kernel.
In addition with other subroutine, can get informations from RDF flux.
For each book in the RDF list flux, extract informations about the book.
Even display it (but should not)

=item B<< get_list_element(flux, key) >>

Can get the value of any node in a flux <ol:key>value</ol:key>.
Can get URI, name and description.
Take the flux and a key as arguments.

=item B<< get_books_number_inside_list(flux_to_parse) >>

Calculate number of "seeds" into our RDF flux.
Take the flux as an argument.

=item B<< get_book_request_url(flux, counter) >>

Get each seed-books URI when called from a loop. Takes the flux contenaing the 
books URIs and a counter as arguments.

=item B<< get_book_element(element) >>

Dynamicaly asks for the element searched into a RDF book flux stored into a file. 
Possible values title, publisher, issued, language. Many other ones are available 
but are not implemented.

=item B<< get_book_id >>

From an URI passed as an argument this subroutine returns the "basename" of the 
URI that is the book id. http://openlibrary.org/books/this->OL24647731M.rdf<-part

=back

=head1 ARGUMENTS

=over 4

=item B<< $flux_to_store >>

The flux himself, directly comming from Open Library.

=item B<< $flux_id >>

The book or list id that identifies this object. Be careful, it's very important 
to be in the good repository ('/database/books/' or '/database/lists/').

=item B<< $list_or_book_selector >>

This is a flag used to tell wich kind of RDF flux we want to treat into the 
Data::RDF::RDFTool


=back

=head1 BUGS AND LIMITATIONS

=over 4

=item IO

Should manage undesirable behaviours just like no-file, no-directory, etc. when 
manipulate directory and files.

=item Conception

Surely a problem in conception of parse_RDF_file() and list / books treatment in the
same subroutine.

=item Authors

If the book have many authors, the authors are stored into an array, not a hash.
See https://bugs.launchpad.net/bibliothek-project/+bug/858280

=back

=head1 AUTHOR

Self Shelf Swap - Smonff - 2011-05-19

=head1 CURRENT MAINTAINER

    my ( $firstname = "smonff", $provider = "lavabit"); 
    print $firstname . "@" . $provider . "." .com;   

=head1 VERSION

This is version 0.02, well known as a Dime And A Shoestring II

=head1 COPYRIGHT AND LICENSE  

Copyright 2011 Self Shelf Swap.

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl himself.

=cut
