# Documentation at the __END__
package Maps::OpenStreetMaps;
#!/usr/bin/perl -w

use strict;
use warnings;
use Data::Dumper;
use version; our $VERSION = qv('0.2.1');
use Data::Properties::PropertiesReader;
use feature qw(say);
use Text::Buffer;

# Return HTML code for inserting a contextual map
# If we are in a lib context, it will build only a single library map
# If we are in a home / main context, it will build a map with all the libraries
# JS must be inserted here 
sub build_map{
	my($self) = @_;
	my $map;
	my($latitude, $longitude, $library_nick, $coordinates);

	# Many libraries map
	if($microlibrary::context{home_page_flag} eq "home" ){
		if ( $microlibrary::log_mode ){
 			say "--- OSM --- Home map = many libraries";
 		}
		# This is a full hash with library names and coordinates
		my %coordinates = &get_coordinates('all');
		$coordinates{location} = "home";
		
		if ( $microlibrary::log_mode ){
			# Loop on all hash values
			foreach my $key (keys %coordinates){
				my $value = $coordinates{$key};
				say "got : " . $value ;
			}
		}

		my $coordinates_ref = \%coordinates;
		$map = &build_JS( $coordinates_ref );

	}

	# One library map
	elsif($microlibrary::context{home_page_flag} eq "lib" ){
		if ( $microlibrary::log_mode ){
 			say "--- OSM --- Lib map = one lib";
 		}

 		$library_nick = $microlibrary::context{nick};
		$coordinates = &get_coordinates( $library_nick );
		
		if ( $microlibrary::log_mode ){
			say '&build_map() got coordinates from &get_coordinates() for $library_nick: ' . "$coordinates";
 		}


 		my %single_coordinates;
 		$single_coordinates{library_nick} = $library_nick;
 		$single_coordinates{coordinates} = $coordinates;
 		$single_coordinates{location} = 'library';
 		
		if ( $microlibrary::log_mode ){
	 		say "Single coordinates";
	 		say Dumper %single_coordinates;
	 	}

	 	my $single_coordinates_ref = \%single_coordinates;
	 	# should pass a reference here
	 	$map = &build_JS( $single_coordinates_ref );

		my $JS_buffer = new Text::Buffer();
		$JS_buffer->append( $map );
 	}

 	# No map (contact page, etc.)
 	elsif( $microlibrary::context{home_page_flag} eq "content"  ){
 		if ( $microlibrary::log_mode ){
 			say "--- OSM --- No map = no library";
 		}
 	}


 	if ( $microlibrary::log_mode ){
 		#say $map;
 	}

 	return $map;
}

sub get_nicks{
	
}

# Get latitude and longitude from the properties file 
# Could also make a get coordinates from outside of the class
sub get_coordinates{
	my $library_name = shift;

	if ( $microlibrary::log_mode ){
		say "Get Open Street Maps coordinates for $library_name library";
	}
	
	my $maps_properties = Data::Properties::PropertiesReader->new();
	my $lat_long_ref = $maps_properties->get_properties_file("maps");

	if ( $microlibrary::log_mode ){
		print Dumper $lat_long_ref;
	}

	# Main map (ALL libraries)
	if( $library_name eq "all"){
		my %extracted_coordinates;
		if ( $microlibrary::log_mode ){
			say "Get coordinates for $library_name library";
		}

		%extracted_coordinates = %{$lat_long_ref};
		if ( $microlibrary::log_mode ){
			say Dumper %extracted_coordinates;
		}

		return %extracted_coordinates;
	}

	# Mono library map. Think to tune the zoom
	else{
		my $extracted_coordinates;
		if ( $microlibrary::log_mode ){
			say "Get coordinates for $library_name";
		}
		$extracted_coordinates = $$lat_long_ref{$library_name};
		if ( $microlibrary::log_mode ){
			say "$library_name extracted coordinates = $extracted_coordinates";
		}
		return $extracted_coordinates;
	}

}


# Pour les mono libraries maps on passe le bordel dans un hash avec un seul couple clé valeur
# Pour la map avec toutes les librairies, on passe direct le hash avec tout
sub build_JS{
	# Is a reference on the hash
	my $transcoordinates_ref = shift;
	my %transcoordinates = %{$transcoordinates_ref};

	my $zoom = 13;
	# DEBUG
	if ( $microlibrary::log_mode ){
		say "Inside build_JS";
		say Dumper %transcoordinates;
	}

	my $JS_map_initialisation = 
		"var map = new L.Map('map');\n".
		"var cloudmadeUrl = 
			'http://{s}.tile.cloudmade.com/45b6834087d44a2a893cd1a55ca725a5/997/256/{z}/{x}/{y}.png', 
			cloudmadeAttrib = '&copy; 2011 + <a href=\"http://www.osmfoundation.org/\">OpenStreetMap</a>', 
			cloudmade = new L.TileLayer(cloudmadeUrl, {maxZoom: 20, attribution: cloudmadeAttrib});\n";

	
	# If we pass a special flag into the hash so we put values in scalars 
	my $JS_map_marker;
	my $JS_map_position;
	if ( $transcoordinates{location} eq 'lib'){
		my $library = $transcoordinates{library_nick};
		my $coordinates = $transcoordinates{coordinates};
	
		if ( $microlibrary::log_mode ){
			say $library;
			say $coordinates;
		}

		$JS_map_position = 
		"var "."$library"."physicalLocation = new L.LatLng("."$coordinates".");\n". 
		"map.setView("."$library"."physicalLocation, "."$zoom".").addLayer(cloudmade);\n";

		# # if scalar, don't loop
		$JS_map_marker = 
		"var markerLocation = new L.LatLng("."$coordinates".");\n".
		"var "."$library"."Marker = new L.Marker(markerLocation);\n".
		"map.addLayer("."$library"."Marker);\n";
	} 
	  else {

		# if number of keys are up to one, so we are going to loop on the hash
		# during the building of JS	
		my $tmp_JS_map_marker;
		my $markers_buffer = new Text::Buffer();;
		# Arbitrary id for explicit decoration into JS
		my $library = "allLibs";
		# All the planet
		$zoom = 2;
		my $coordinates = "51, -45";
		$JS_map_position = 
		"var "."$library"."physicalLocation = new L.LatLng("."$coordinates".");\n". 
		"map.setView("."$library"."physicalLocation, "."$zoom".").addLayer(cloudmade);\n";

		# foreach my $key...
		foreach my $library (keys %transcoordinates){
			my $coordinates = $transcoordinates{$library};
			next if $coordinates eq 'home';
			if ( $microlibrary::log_mode ){ 
				say "got : " . $coordinates ;
			}
			$tmp_JS_map_marker = 
				"var markerLocation = new L.LatLng("."$coordinates".");\n".
				"var "."$library"."Marker = new L.Marker(markerLocation);\n".
				"map.addLayer("."$library"."Marker);\n";

			# pour chaque couple clé valeur, on rajoute le JS au buffer
			$markers_buffer->append( $tmp_JS_map_marker );

			
		}
		if ( $microlibrary::log_mode ){
			say "Markers : ";
			say $markers_buffer->dumpAsString();
		}
		$JS_map_marker = $markers_buffer->dumpAsString();

	}



	my $returning_map_code = "$JS_map_initialisation" . "$JS_map_position" . "$JS_map_marker";
	#say $returning_map_code and die;

	return $returning_map_code;

}


sub new {
    my $class = shift;
    my $self = {};
    bless ($self, $class);
    if ( $microlibrary::log_mode ){
    	say "- New OpenStreetMap";
    }
    return $self;
    # DEBUG
}

1;

__END__

=pod

=head1 NAME

Maps::OpenStreetMaps - Leaflet API Open Street Map implementation for Self Shelf Swap

=head1 DESCRIPTION

Provides a way of getting latitude & longitude coordinates from a properties file and
display it into XHTML pages.

=head1 VERSION

This is version 0.2.1, better known as a Chines Loony And A Rope
(the third Dime And A Shoestring)

=head1 CONFIGURATION AND ENVIRONMENT

There is a special Properties file (<<maps.properties>>) that has the library id
as a key and latitude- longitude duo like a value

=head1 SUBROUTINES / METHODS

=over 4

=item B<< >>

xxxxxxxxxxxxxxxxxxxxxxxx

=back

=head1 ARGUMENTS

Could be specified, depending on what subroutine you will use.

=over 4

=item B<< Tag >>

xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

=back

=head1 AUTHOR 

Self Shelf Swap - Smonff - 2012-02-19

=head1 CURRENT MAINTAINER

    my ( $nickname = "smonff", $provider = "lavabit.com");
    print $nickname . "@" . $provider;

=head1 COPYRIGHT AND LICENSE

Copyright 2012 Self Shelf Swap.

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl himself.

=cut
