# Documentation at the __END__
package View::WebDisplayer;
#!/usr/bin/perl -w

use strict;
use warnings;
use Data::Dumper;
use version; our $VERSION = qv('0.2.1');
use Template;
use feature qw(say);
use Perl6::Slurp;
use Data::Properties::PropertiesReader;
use Text::Buffer;

my $microlibrary_name;
my $microlibraries_URLs_list ="";

# # Template for header
# my $template_head_file = 'head.tmpl';
# my $head_template = Template->new({
#   # Mutli path is possible with ':' separating
#     INCLUDE_PATH => '/View/tmpl',
#     PRE_PROCESS  => 'config',
# });


# Go away with V0.2.1!!!
my $header_menu = '<div id="head">
    <div id="header-inn">
      <a href="http://selfshelfswap.org/wiki/">Wiki</a>
      <a title="Work with us!" href="https://launchpad.net/bibliothek-project">LaunchPad</a>
      <a title="Report a bug" href="https://bugs.launchpad.net/bibliothek-project">Bugs</a>
    </div>
    </div>
    <div id="question">
        <a 
         title="If you want to ask us a question, please fill this form, we will answer you soon." 
         href="./mail/ContactForm/contact.pl">Question ?
      </a>
    </div>
    <div id="donate">
      
      <form id="paypal" action="https://www.paypal.com/cgi-bin/webscr" method="post">
        <input type="hidden" name="cmd" value="_s-xclick">
        <input type="hidden" name="hosted_button_id" value="MHHS7UXHJS2H6">
        <input type="image" src="https://www.paypalobjects.com/en_US/i/btn/btn_donate_LG.gif" border="0" name="submit" alt="PayPal - The safer, easier way to pay online!">
        <img alt="" border="0" src="https://www.paypalobjects.com/fr_FR/i/scr/pixel.gif" width="1" height="1">
      </form>
      <a href="http://selfshelfswap.org/wiki/index.php/Self_Shelf_Swap_Community_Wiki:Site_support" class="donate_button"><!--<img src="./View/style/img/money_yesl.png" alt="money logos"/>--><br />Sssupport us !</a>
    </div>';



my $logo = '<div id="logo-principal">
    <img alt="Self shelf swap\'s logo : 3 colored snake-shelves" src="/View/style/img/selfShelfSwap.png" />
</div>';

#my $footer = '</html>';

my $js_libs = '<div id="scripts">
                 <script type="text/javascript" src="./View/js/scroller.js"></script>
               </div>';

# Used to put open XHTML tags that will be close soon
# TODO V0.2 : Force to close the tag or warn if it isn't
# TODO V0.2 : Indent tags
my @tag_stack;
# What each subroutine will print
my $render;

# Prints content into XHTML tags
sub print_into_tags{
  my ( $self, 
         $content, 
         $tag, 
         $param_key, 
         $param_value, 
         $param_key2, 
         $param_value2 
      ) = @_;

  # With class and id
  if ( $param_key && $param_key2){
        $render = "<" . $tag . " " . 
                        $param_key . "=\"" . "$param_value\" " . 
                        $param_key2 . "=\"" . "$param_value2\"" . ">" . 
                        $content .
                         "</" . 
                        $tag . ">";
    }

  # With class or id
  elsif ( $param_key ){
    $render = "<" . $tag . " " . 
                    $param_key . "=\"" . "$param_value\"" . ">" . 
                    $content . "</" . 
                    $tag . ">";
#        if ( $microlibrary::log_mode eq '1'){
#        print $render;
#   }
  }

  # Tag only
  else{
    $render = "<" . 
    $tag . ">" . 
    $content . "</" . 
    $tag . ">";
  }

  $microlibrary::html_buffer->append( $render );
}

# Open-only tag (bad human closing)
sub open_tag{
  my ( $self, $tag, $param_key, $param_value, $param_key2, $param_value2 ) = @_;

  #TODO V0.2 : Force to close the tag or warn if it isn't
  push( @tag_stack, $tag );
  #print Dumper @tag_stack;

  # Class and id
  if ( $param_key && $param_key2 ){
    $render = "<" . $tag . " " . $param_key . "=\"" . "$param_value\"" . " " . $param_key2 . "=\"" . "$param_value2\"" . ">";
  }

  #Class or id
  elsif( $param_key ){
    $render = "<" . $tag . " " . $param_key . "=\"" . "$param_value\"" . ">";
  }

  # Tag only
  else{
    $render = "<" . $tag . ">";
  }

  $microlibrary::html_buffer->append( $render );
}

# Automatised close-only tag (after human opening)
sub close_tag{
    my ( $self, $newline, $end, $end_value ) = @_;

    # Check for last opened tag
    my $closing_tag = pop (@tag_stack);

    if ( defined ( $end ) && defined ( $end_value ) ){
      my $end_comment = "<!-- End " . $end_value . " -->";
      $microlibrary::html_buffer->append( $end_comment );
    }

    $render = "</" . $closing_tag . ">";
    $microlibrary::html_buffer->append( $render );

    # Can add a backchar
    if ( defined ( $newline ) && ( $newline eq "newline" )){
      &newline;
    }
}

# Includes a .tmpl page into the HTML buffer
# It's not necessary to give the extension of the file into the argument
sub include_tmpl{
  my ( $self, $template ) = @_;

  # TODO : Test de l'existence du template !!!!!!

 $render = "[% INCLUDE " . $template . ".tmpl %]\n";
 $microlibrary::html_buffer->append( $render );

}

# Put the processed template from a temporary buffer to the $view html_buffer
sub append_tmp_tmpl{
  my ($self, $template_buffer_name) = @_;
  my $template_file_handle = FileHandle->new;
  my $flag;

  if ( $microlibrary::log_mode ){
    say "FLAG : " . $template_buffer_name;
  }

  if(defined $template_buffer_name){
    $flag = $template_buffer_name;
  } else {
    $flag = "tmp_template_buffer";
  }

   if ( $microlibrary::log_mode ){
    say "FLAG : " . $flag;
  }

  my $tmp_path = $microlibrary::tmp_dir . "/" .$flag;

  if ($template_file_handle->open("< $tmp_path")) {
    if ( $microlibrary::log_mode ){
      say "File open";
    }
    my $slurped_file = slurp($template_file_handle);
    # say Dumper $slurped_file;
    return $slurped_file;
  }
  else{
    die("Cannot ssslurp processed template file");
  }
}

# Obsolete in V0.02.1 
# # Inserting INCLUDE
# sub header{

#     my $header = "[% INCLUDE head.tmpl%]";
#     $microlibrary::html_buffer->append($header . "\n");
# }

sub header_menu{
  $microlibrary::html_buffer->append( $header_menu . "\n" );
}

sub logo{
  $microlibrary::html_buffer->append( $logo . "\n" );
}

sub make_sidebar_menu{

}

sub sidebar_menu{ # BUG !!!
                  # Never put the array before the hash reference, it makes shit
                  # (hash will be INTO the array)
                  #say "#"x 100;
                  #say Dumper @_ and die;
  my ($self, @nicks_list  ) = @_; #$home_flag
  # WAS my ( $self, @nicks_list,  ) = @_; #$home_flag
  if ( $microlibrary::log_mode ){
        print "Into sidebar menu !!!";
        #print Dumper @nicks_list;
        
        # say $$library_vars{library}{name};
        # say $$library_vars{library}{location};
        # say $$library_vars{library}{OL_account};
        #my %hash_test = %{$library_vars};
        # say Dumper %hash_test and die;
        #my $test = $$library_vars{'library'}{'name'};
        #say "TEST : " . $test and die;
        #print $home_flag;
        #die;
    }

    ## GROS PROBLEME DE DECALAGE ENTRE LA VALEUR DANS $$library_vars ET DANS @nicks_list
    # Faire une routine get_library_vars() pour récupérer la valeur depuis l'itération
    # sur l'array de noms de bibliotheques

    my $list_html_buffer = Text::Buffer->new();

    for my $nick ( @nicks_list ){

        # breaks the loop when nothing has to been found into the array

        # CORRECTED BUG !
        # if ( $nick eq 1 ){ last; }
        if ( $nick eq 0 ){ last; }

        if ( $microlibrary::log_mode ){
          print "Passed into the loop ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; \n";
          print "$nick\n";
        }

        $microlibrary_name = &get_microlibrary_name( $nick );

        if( $microlibrary_name && $microlibrary_name ne "home" ){

            if ( $microlibrary::log_mode ){
              print $microlibrary_name . "\n";
            }

            # Flag for the menu to generate a active or not class
            my $active_or_not;
            if( $nick eq $microlibrary::context{nick} ){
              $active_or_not = "active";
            } else {
              $active_or_not = "passive";
            }


            # Get the data from properties file
            my $modes_properties = Data::Properties::PropertiesReader->new();
            my $libraries_infos_ref = $modes_properties->get_properties_file("libraries_infos");
          
            my $microlibrary_menu_vars = {
                # Library infos
                menu        => {
                  status      => $active_or_not,
                  path        => "?lib=".$nick,
                },
                library   => {
                    name              =>  $$libraries_infos_ref{$nick}{name},
                    location          =>  $$libraries_infos_ref{$nick}{location},
                    OL_account        =>  $nick,
                },
            };


            if ( $microlibrary::log_mode ){
              say Dumper $microlibrary_menu_vars;
            }

            # SHOULD BE MUTUALIZED !!!!!
            # Ajouter les éléments de list à un buffer dans un $calar ($list_html_buffer) 
            # et enfin les injecter dans le template principal (en interraction avec microlibrary.pl)
            my $menu_elements_template = Template->new($microlibrary::templates_conf);
  
            my $template_menu_elements_file = 'list_elements.tmpl';
            my $tmp_template_buffer = '';
  
            $menu_elements_template->process($template_menu_elements_file, $microlibrary_menu_vars, 'list_elements' )
              || die $menu_elements_template->error("Can't process template menu :(");
            
            # OLD STYLE
           
            # $microlibraries_URLs_list = $microlibraries_URLs_list . 
            #     "<li>
            #         <a href=\"http://$microlibrary::host/microlibrary.pl?lib=$nick\" class=\"$active_or_not\">
            #             " . $microlibrary_name  . " library
            #         </a>
            #     </li>\n";


        }

        my $template_file_handle = FileHandle->new;
        if ($template_file_handle->open("< $microlibrary::tmp_dir/list_elements")) {
          if ( $microlibrary::log_mode ){
            say "File open";
          }
          my $slurped_file = slurp($template_file_handle);
          # say Dumper $slurped_file;
          # Concatenate HTML tags with dynamical values (not replace)
          $list_html_buffer->append($slurped_file);
        }
        else{
          die("Cannot ssslurp processed template file");
        }

    }

  
  # if ( $microlibrary::log_mode ){
  #   say $list_html_buffer->dumpAsString();
  # }
 
  # Returns the Text::Buffer object as a string
  return $list_html_buffer->dumpAsString();
}

# V0.2.1 OBSOLETE !!!
# Hardcoded shit !!!!
# Should be dynamically into properties or DB
sub get_microlibrary_name{
  my $OLnick = shift;

  if ( $microlibrary::log_mode ){
      print "Into get_library_name()\n";
        print $OLnick . "\n";
  }

  if( $OLnick ne "home" && $OLnick ne "lib" ){

    if ($OLnick eq "sandoth" ){
      $microlibrary_name = "The STAG";
    }

    elsif ($OLnick eq "carle_colline" ){
        $microlibrary_name = "Carle Colline";
    }

    elsif ($OLnick eq "simonfoutaiz" ){
        $microlibrary_name = "S. Foutaiz";
    }

    elsif ($OLnick eq "default" ){
        $microlibrary_name = "Default guy";
    }

    elsif ($OLnick eq "Fizzyball" ){
        $microlibrary_name = "Fizzyball";
    }
    
    elsif ($OLnick eq "Noil" ){
        $microlibrary_name = "Noil";
    }

    elsif ($OLnick eq "SteamShip" ){
      $microlibrary_name = "SteamShip";
    }

    elsif ($OLnick eq "yeyetoon" ){
      $microlibrary_name = "Yéyé Toon";
    }

    elsif ($OLnick eq "ZooeyGlass" ){
      $microlibrary_name = "Zooey Glass";
    }

    else{
        die "Not captured microlibrary name : 
              maybe it's an error, maybe you should add a case into View::WebDiplayer"; 
    }
  }
    return $microlibrary_name;

}


sub just_fucking_print{
  my($self, $damned) = @_;
  $microlibrary::html_buffer->append( $damned );
}

sub append{
  my($self, $stuff) = @_;
  $microlibrary::html_buffer->append( $stuff );
}

# TODO V0.2 : should be automatic, should read tags and understand wich level of indentation is necessary
sub newline{
  $microlibrary::html_buffer->append( "\n" );
}

# Obsolete in V0.02.1 
# sub close_footer{
#   $microlibrary::html_buffer->append( $footer );
# }

sub insert_JS{
  $microlibrary::html_buffer->append( $js_libs );
}

# TODO V0.2 : should be automatic, should read tags and understand wich level of indentation is necessary
sub indent{
}
# TODO V0.2 : should be automatic, should read tags and understand wich level of indentation is necessary
sub unindent{
}


sub new {
    my $class = shift;
    my $self = {};
    bless ($self, $class);
    if ( $microlibrary::log_mode ){
      say "- New WebDisplayer";
    }
    return $self;
    # DEBUG
}

if ( $microlibrary::sad_deps_mode ){
    my $env_viewer = Tools::DependanciesViewer->new();
    $env_viewer->print_used_modules($0, %INC);
}

1;

__END__

=pod

=head1 NAME

View::WebDisplayer - Microlibrary renderer for repetitive print stuffs

=head1 DESCRIPTION

SAD (Simple And Dirty) script for web generating XHTML tags. Can be used for
templates, by printing XHTML tags into your Microlibrary web-application output .
No MVC at all (we were in a hurry). Just instanciate a new View::WebDisplayer and
you are ready to generate tags.  Depending on wich subroutine you use,
you can pass parameters, see ARGUMENTS section.

=head1 VERSION

This is version 0.02.1, well known as a the Chinese Loony And A Rope

=head1 CONFIGURATION AND ENVIRONMENT

There is some harcoded XHTML things on the top of the file.
Feel free to remplace the code into C<< $header >>, C<< $header_menu >>,
C<< $sidebar_menu >>, C<< $logo >> and C<< $footer >> to generate a different website.
This can be used to generate a very low-tech template. Names are well choosen that's
our great pride!

=head1 SUBROUTINES / METHODS

=over 4

=item B<< print_into_tags() >>

Open & close tags with many parameters. Should use it when you don't want to
imbricate something, especially for inline tags like <span></span> for example.
    EXAMPLE : <tag key="value">content of the tag</tag>

=item B<< open_tag() >>

Open the tag you want with a single, many or none parameter(s).
Automatically put the tag into the C<< @tag_stack >>, that will be emptied when
you will close this trash-tag.

=item B<< close_tag() >>

Close the last tag you have open. Should not exist, should close automatically.
Can take a C<< $newline >> argument.

=item B<< header() >>

Just print the XHTML header you've specified into C<< $header >>.

=item B<< header_menu() >>

Just print the XHTML header menu you've specified into C<< $header_menu >>.

=item B<< logo() >>

Just corporate-print it !

=item B<< header() >>

Just print it !

=item B<< just_fuckin_print() >>

Just fuckin' print it ! Takes a damned scalar in argument.

=item B<< append() >>

Just fuckin' print it ! Takes a stuffed scalar in argument. Makes exactly the 
same thing than just_fuckin_print(), but cite the Text:Buffer syntax.


=back

=head1 ARGUMENTS

Could be specified, depending on what subroutine you will use.

=over 4

=item B<< Tag >>

Name of a valid XHTML tag

=item B<< Content >>

Whatever that will be inserted into the tag. You can pass scalars or static content.

=item B<< Keys and values >>

One, two or none key-values peers for class="" and/or id="" parameters.

=item B<< Newline >>

Add a backchar for clean XHTML code. Next year we will code something for indentation.


=back

=head1 AUTHOR 

Self Shelf Swap - Smonff - 2011-06-29

=head1 CURRENT MAINTAINER

    my ( $nickname = "smonff", $provider = "lavabit.com");
    print $nickname . "@" . $provider;

=head1 COPYRIGHT AND LICENSE

Copyright 2012 Self Shelf Swap.

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl himself.

=cut
