#!/usr/bin/perl -w
package microlibrary;

# Documentation at the __END__
use strict;
use warnings;
use feature qw(say);
use Data::Dumper;
use Data::Properties::PropertiesReader;
use Data::CURL::CurlRequester;
use Data::RDF::RDFTool;
use View::WebDisplayer;
use Tools::DependanciesViewer;
# use Module::ScanDeps;
use CGI;
use Text::Buffer;
use FileHandle;
#use CGI::Carp;
use File::Basename;
use Properties;
# use Log::Log4perl;
use Time::HiRes qw(gettimeofday); 
use Maps::OpenStreetMaps;
use version; our $VERSION = qv('0.2.1');
use Template;
use Perl6::Slurp;

# Used to display nice performance stuffs
(my $begin_epoch_seconds, my $begin_epoch_microseconds) = gettimeofday;
(my $second, my $minute, my $hour) = localtime($begin_epoch_seconds);
my $display_beginning_time =  sprintf("%02d:%02d:%02d", $hour, $minute, $second ) . "\n"; 

# Next we will implement log this way, cause we make a very dirty home-made logger ('mode_log')
#Log::Log4perl::init( "../properties/logs.properties" );
#my $log = Log::Log4perl::get_logger("Self::Shelf::Swap");
#$log->logcroak("croak");  

my $delimiter = "-";
my $book_URL;
my $modes_properties = Data::Properties::PropertiesReader->new();
my $modes_ref = $modes_properties ->get_properties_file("modes");

# PROPERTIES LOADER SEPARATE MODULE FOR V0.2.1
# See 'modes.properties' file to set it
# 'Reload mode' will go and search for microlibraries data on the Open Library database.
# Always turn it off when your prod environment is ready.
# True  (string containing "1") reload RDF flux from Open Library, and update data on disk
# False (string containing "0") load lists and books from files on local disk
our $reload_mode = $$modes_ref{mode}{reload};
# Logs to STDOUT if 1
our $log_mode = $$modes_ref{mode}{log};  # For dev env
our $browser_simulation_mode = $$modes_ref{mode}{browser}{simulation};
# Print HTML for the web if 1
our $html_mode = $$modes_ref{mode}{html};  # For prod env
# Scans for dependencies in the SAD way
our $sad_deps_mode = $$modes_ref{mode}{sad};
# Simulates a query string when pages are 'called' from command line
our $browser_debug_mode = $$modes_ref{mode}{browser}{debug};

my $nick = "";

my %hashed_rdf_fluxus_ref;

our @shelves = ();
my $shelves_size = @shelves;

# If you want to have everything print into a browser you should uncomment bellow 
# print("Content-type: text/html\n\n");

if ( $microlibrary::log_mode ){
    print "- SELF SHELF SWAP - Dime And A Shoestring - V0.02 -\n";
    print "--Mimicrokernel started at $display_beginning_time\n";
}

my $env_properties = Data::Properties::PropertiesReader->new();
my $env_ref = $env_properties ->get_properties_file("env");

my $env = $$env_ref{env}{home};
our $host = $$env_ref{env}{host};
our $tmp_dir = $$env_ref{env}{tmp};

if ( $microlibrary::log_mode ){
    say "tmp directory = " . $tmp_dir;
}

# Debug
#print(Dumper $env);


#"/this/is/the/path/to/your/dev/app/lib/" or
#"/the/path/to/your/prod/app/lib"
$ENV{HOME} = $env;

# Check that the path for $ENV{HOME} directory exists
if ( $microlibrary::log_mode ){
    # -directory
    if (-d "$ENV{HOME}/") {
            print "--- ENV{HOME} set OK !!\n";
    } else {
        die ('-- !!! FATAL Bad $ENV{HOME} path please set it !!!') ;    
    }

    # Chez that this script is well located into the $ENV{HOME}
    my $script_name = basename($0);
    # DEBUG
    #print "$ENV{HOME}" . '/' . "$script_name";
    if ( ! -e $ENV{HOME} . '/' . $script_name ){
        die ("-- !!! FATAL $script_name is not located into " . 'ENV{HOME} !!!\n') ;;
    } else {
        print "--- THIS SCRIPT IS LOCATED INTO ENV{HOME} DIRECTORY !!\n"
    }
}

# A nice flag control panel into console
if ( $microlibrary::log_mode ){
    print(
        "      * * * * * * * * * *\n", 
        "      *                 *\n", 
        "   M  *    RELOAD = $reload_mode   *\n", 
        "   O  *    LOGS = $log_mode     *\n", 
        "   D  *    HTML = $html_mode     *\n",
        "   E  *    SIMU = $browser_simulation_mode     *\n",
        "   S  *    SADS = $sad_deps_mode     *\n",
        "      *    DBG = $browser_debug_mode      *\n",
        "      *                 *\n",
        "      * * * * * * * * * *\n"
    );
}


# Goin' to search for the value of any parameter called 'lib' into an existing query string
my $cgi = CGI->new();
my $query_string_param_ref = $cgi->Vars;
 
# If 'home' is asked, display home-page, else that's a library page have been asked
my $home_page_flag = "home";

if( defined $query_string_param_ref ){
#     say "toto";
    if( defined $$query_string_param_ref{lib} ){
        # We activate the lib flag)
        $home_page_flag = "lib";

        if ( $microlibrary::log_mode ){ 
            print Dumper $$query_string_param_ref{lib};
        }
        $nick = $$query_string_param_ref{lib};
        if ( $microlibrary::log_mode ){ 
            print "---- QUERRY STRING = ?lib=$nick \n";
        }
    }

    elsif ( defined $$query_string_param_ref{page} ){
        if ($$query_string_param_ref{page} eq "contact"){
            $home_page_flag = "contact";    
        } elsif ($$query_string_param_ref{page} eq "about"){
            $home_page_flag = "about"; 
    } 

    else{
       $home_page_flag = "fuck"; 
    }

} else {

    if ( $microlibrary::log_mode ){ 
        print "---- QUERRY STRING = UNDEF \n";
        $home_page_flag = "home";
    }

    if ( $microlibrary::browser_simulation_mode == 1 ){
        # Trick for simulate the query-string from the browser when working into console
        $nick = "sandoth"; # simonfoutaiz sandoth  ZooeyGlass    yeyetoon   carle_colline  SteamShip  Fizzyball    Noil
        if ( $microlibrary::log_mode ){ 
            print "---- SIMULATED QUERRY STRING ON = ?lib=$nick \n";
        }
        $home_page_flag = "lib";
    }

    # No querry string, no simulation-mode, we ask for the home
    elsif ( $microlibrary::html_mode == 1 ){
        $nick = "home";
        $home_page_flag = "home";
        if ( $microlibrary::log_mode ){ 
            print "---- This is home page \n";
            print "----- home_page_flag = $home_page_flag\n";
        }
    }
 }

# On va chercher les id des lists Open Library dans un fichier de propriétés
my $OL_prop_reader = Data::Properties::PropertiesReader->new();
my $my_file_content = $OL_prop_reader->get_properties_file("open_library_lists");

if ( $microlibrary::log_mode ){
    print Dumper $my_file_content;
}

# Because I don't want to pass parameters about this flags and return it
# into arrays where they should not be !!!!
# This way everything is encapsulated into only one hash and it's better
# We could put the "modes" inside too
our %context = (
    'home_page_flag' => $home_page_flag,
    'nick' => $nick,
);

# Can be used for generate menu and querry parameters
# Nicks will be give to the view to generate URLs for the side-menu
my @OL_users_keys = $OL_prop_reader->get_nicks_keys( $my_file_content ); # Removed the horrible home_page_flag here
if ( $microlibrary::log_mode ){
    print "/////////// Back to microlibrary !\n";
    print Dumper(@OL_users_keys);
} 

my @OL_list_ids = $OL_prop_reader->get_list_ids( $my_file_content, $nick ); # Remove this horrible flag as soon as possible
if ( $microlibrary::log_mode ){
    print "/////////// Back to microlibrary !\n";
    print Dumper(@OL_list_ids);
}

# Puts a contextual map into cache
# It will be called later as a static file
my $map = Maps::OpenStreetMaps->new();
# Don't need to return anything, only for debug
# !!!!!!! DEBUG !!!!!
my $JS_map_code = $map->build_map();
# DEBUG
#say $JS_map_code and die;

# Not finshed , not used
# if ( $microlibrary::log_mode ){
#     my $map_code = FileHandle->new;
#     if ($map_code->open("< ./View/js/OpenStreetMaps.js")) {
#         print <$map_code>;
#         $map_code->close;
#     } else {
#         die("Oups, the file containing cached html don't exists...");
#     }
# }

# Display beginning
our $html_buffer = new Text::Buffer();
my $view = View::WebDisplayer->new();
# General templates configuration stuff
# See http://search.cpan.org/~abw/Template-Toolkit-2.22/lib/Template.pm#___top
# Config is stored into a single hash, that can be used my many templates
our $templates_conf = {
    RELATIVE        => 1,
    # Mutli path is possible with ':' separating
    INCLUDE_PATH    => './View/tmpl',
    PRE_PROCESS     => './View/tmpl/config',
    OUTPUT_PATH     => './View/tmpl',
    #OUTPUT_PATH     => '/tmp/selfshelfswap',
};

if ( $microlibrary::html_mode ){
    # Expiration is set into Apache configuration
    # $cgi->header(-charset => 'utf-8', -expires => '+3d');
    $cgi->header(-charset => 'utf-8');
    # To test too
    # $cgi->header(-type=>"text/html; charset=utf-8")
    print $cgi->header();
    # Old bugged style stuff
    #print "Content-type: text/html\n\n";
    #print("Content-type: text/html");
    #$html_buffer->append("Content-type: text/html"); # \n\n Weird bug : sometimes you will need to comment this pragma  and use the one above
    
    # <head>

    #say Dumper $query_string_param_ref;

    $view->include_tmpl("head");
    $view->open_tag( "body", "id", "$context{'home_page_flag'}" );
    $view->newline();
    $view->open_tag( "div", "class", "site" );
    $view->newline();

    $view->open_tag( "div", "id", "page" );
    $view->newline();

    $view->include_tmpl("header");

    $view->open_tag( "div", "id", "main", "role", "main" );
    $view->newline();
    if (  $microlibrary::log_mode  ){
        print "home_page_flag : $home_page_flag\n";
    }

    $view->open_tag( "div", "class", "container",  );

}


if ( $context{'home_page_flag'} eq "home" && $OL_list_ids[1] eq "sweet"){ # If we ask for home page, @OL_list_ids = home, sweet, home :)

    # &generate_home_page();
    if( $microlibrary::log_mode ){
        print " We built the standard HOME PAGE "; # On fait la home page de base
    }

    if ( $microlibrary::html_mode ){

        # ALWAYS USE THE VARS, DON'T TRY TO PROCESS WITHOUT VARS
        my $start_vars = {
            page     =>{
                host  => $host,
            },
        };

        my $start_template = Template->new($microlibrary::templates_conf);

        my $template_file = 'get_started.tmpl';
        # say "ok";
        # my $tmp_template_buffer = '';

        # Add ""
        $start_template->process($template_file, $start_vars, 'tmp_template_buffer')
            || die $start_template->error();

        #say "ok";

        # say "Template processed";

        my $processed_template = $view->append_tmp_tmpl();
        $view->append($processed_template);

        $view->close_tag("newline", "comment", "container");
        $view->close_tag("newline", "comment", "main");

        # my $template_file_handle = FileHandle->new;
        # if (my $template_file_handle->open("< ./View/tmpl/tmp_template_buffer")) {
        #     say "File open";
        
        #     my $slurped_file = slurp($template_file_handle);
        #     # say Dumper $slurped_file;
        #     say $slurped_file;
        # }

        # else{
        #     die("Cannot ssslurp processed template file");
        # }
        
        # die;



            $view->close_tag();
            #$view->close_tag();
            #$view->close_tag();
            &print_generic_footer;

            #$view->close_tag();  
    }

#print $context{'home_page_flag'};


# Equal lib
} elsif( $context{'home_page_flag'} eq "lib" )  {

    # Microlibrary informations block
    if ( $microlibrary::html_mode){

        # Get the data from properties file
        my $libraries_infos_ref = $modes_properties->get_properties_file("libraries_infos");
        
        my $OL_account = "http://openlibrary.org/people/" . $context{nick};
        #print $OL_account and die;

        
        if( $microlibrary::log_mode ){
            say Dumper $JS_map_code;
        }

        my $library_vars = {
            # Library infos
            'library'   => {
                'name'              =>  $$libraries_infos_ref{$context{nick}}{name},
                'description'       =>  $$libraries_infos_ref{$context{nick}}{description},
                'description_more'  =>  $$libraries_infos_ref{$context{nick}}{description_more},
                'email'             =>  $$libraries_infos_ref{$context{nick}}{email},
                'website'           =>  $$libraries_infos_ref{$context{nick}}{website},
                'location'          =>  $$libraries_infos_ref{$context{nick}}{location},
                'OL_account'        =>  $OL_account,
                'JS_map_code'       =>  $JS_map_code,
                # Another hash into the hash could be stored ???
                # ...  
            },
        };

        $context{lib_display_name} = $$library_vars{library}{name};

        if ( $microlibrary::log_mode ){
            say Dumper $library_vars;
        }

        # TODO : Change the name of this subroutine, ONLY FOR MICROLIBRARY MODE, not for home 
        # Always put the array at the end of the parameters, or their will be a bug  
        my $microlibraries_list_menu = $view->sidebar_menu(@OL_users_keys ); #$home_page_flag

        # say Dumper $microlibraries_list_menu;

        my $menu_template_vars = {
            menu     =>{
                list        => $microlibraries_list_menu
            },
        };

        my $menu_template = Template->new($microlibrary::templates_conf);

        my $menu_template_file = 'ml_menu.tmpl';
        my $tmp_menu_template_buffer = '';

        $menu_template->process($menu_template_file, $menu_template_vars, 'tmp_ml_menu_buffer')
            || die $menu_template->error("Can't process menu template :(");

        # say "Template processed";

        my $processed_template = $view->append_tmp_tmpl("tmp_ml_menu_buffer");
        $view->append($processed_template);


#        $view->append($microlibraries_menu);

        my $library_infos_template = Template->new($templates_conf);

        my $info_template = 'ml_info.tmpl';
        my $tmp_template_buffer = '';

        # Processing the template (arguments are template file, vars and output file relative to OUTPUT_PATH value)...
        # my $microlibrary_infos = 
        $library_infos_template->process($info_template, $library_vars, 'tmp_ml_infos_template_buffer')
            || die $library_infos_template->error("Can't process microlibrary infos template :(");


        my $processed_ml_infos_template = $view->append_tmp_tmpl('tmp_ml_infos_template_buffer');
        $view->append($processed_ml_infos_template);

        # Now add the tmp_template_buffer to the $html_buffer
        # has been MUTUALIZED INTO WebDisplayer.pm
        # my $template_file_handle = FileHandle->new;
        # if ($template_file_handle->open("< ./View/tmpl/tmp_template_buffer")) {
        #         if ( $microlibrary::log_mode ){
        #             say "File open";
        #         }
        #         my $slurped_file = slurp($template_file_handle);
        #         $view->append($slurped_file);
        # }
    }

    if ( $reload_mode ){

        my $request_beginning_time = localtime();
        # On passe les ids au CurlRequester afin de recevoir une requete
        my $list_data_from_OL = Data::CURL::CurlRequester->new();
        if ( $microlibrary::log_mode ){
            print Dumper @OL_list_ids;
        }

        %hashed_rdf_fluxus_ref = $list_data_from_OL->build_the_list_request( $nick, @OL_list_ids );
        my $request_ending_time =  localtime();
        if ( $microlibrary::log_mode ){
            print "STATS :\n    Restful API requests beginning time : $request_beginning_time\n";
            print "    Restful API requests beginning time :Ending process : $request_ending_time\n";
        }
    }

    if ( $reload_mode ){

        my $rdf_flux_writer = Data::RDF::RDFTool->new();
        my $rdf_flux_accessor;

        # Store each RDF books list flux in our hash for every id stored in the array
        # ( hash keys are the same than array values )
        if ( $microlibrary::log_mode ){
            print "\n-- Data serialization " . $delimiter x 50 . "\n";
        }
        # Send what we get from CurlRequester to RDFTool
        foreach ( @OL_list_ids ){
            $rdf_flux_writer->store_rdf_flux(
                $hashed_rdf_fluxus_ref{$_}, $_, "list"    # Last parameter is a flag : asks for a list or a book
            );
        }
    }

    # Read, parse the rdf flux and cache it on disk into files
    #print "\n-- Serialized data reader and parser " . $delimiter x 50 . "\n";
    my $rdf_file_containing_books_list = Data::RDF::RDFTool->new();
    my $OL_list_idx=0;
    my $book_data_from_OL;

    #print Dumper $microlibrary::context{main_list_description};
    #die;

   # TODO : Hard to add the description cause it it not initialized here
   # See code in RDFTool too
   # my $description = $microlibrary::context{main_list_description};

    if ( $microlibrary::html_mode ){
        $view->open_tag( "div", "id", "bookshelves" );
        $view->newline();
    #    $view->print_into_tags( "$description", "h3");
    }

    # Building the shelves menu
    my $shelves_menu_buffer = Text::Buffer->new();

    my $shelf_idx = 0;

    for my $shelf_id ( @OL_list_ids ){

        my $shelves_menu_vars = {
            # Shelves infos
            shelf   => {
                id      => $shelf_id,
                name    => "[% shelfsize.$shelf_idx.name %]",
            },
        };

        $shelf_idx++;

        # SHOULD BE MUTUALIZED !!!!!
        # Ajouter les éléments de list à un buffer dans un $calar 
        # et enfin les injecter dans le template principal 
        my $shelves_menu_template = Template->new($microlibrary::templates_conf);

        my $shelves_menu_template_file = 'shelf_menu_element.tmpl';
        my $tmp_template_buffer = '';

        $shelves_menu_template->process($shelves_menu_template_file, $shelves_menu_vars, 'tmp_shelves_menu_buffer' )
          || die $shelves_menu_template->error("Can't process shelves template menu :(");


        my $template_file_handle = FileHandle->new;
        if ($template_file_handle->open("< $microlibrary::tmp_dir/tmp_shelves_menu_buffer")) {
          if ( $microlibrary::log_mode ){
            say "File open";
          }
          my $slurped_file = slurp($template_file_handle);
          # say Dumper $slurped_file;
          # Concatenate HTML tags with dynamical values (not replace)
          $shelves_menu_buffer->append($slurped_file);
          if ( $microlibrary::log_mode ){
            say "ADD DATA";
          }
        }
        else{
          die("Cannot ssslurp processed template file");
        }

    }

    my $final_menu_buffer = $shelves_menu_buffer->dumpAsString();
    #say $final_menu_buffer;


     ################### > INTEGRATION of shelfes templates

    my $shelves_vars = {
        # Library infos
        'library' => {
            # Made to display the library name into the shelves.tmpl template
            'name'              =>  $microlibrary::context{lib_display_name},
        },
        'html'    => {
            'code'              =>  $final_menu_buffer,
        },
    };

    ## Process shelf 

    my $shelves_template = Template->new($microlibrary::templates_conf);

    my $shelves_template_file = 'shelves.tmpl';
    my $tmp_template_buffer = '';

    $shelves_template->process($shelves_template_file, $shelves_vars, 'tmp_shelf_template_buffer')
        || die $shelves_template->error("Can't process shelves template :(");

    if ( $microlibrary::log_mode ){
        say "Template processed";
    }



    my $processed_shelves_template = $view->append_tmp_tmpl('tmp_shelf_template_buffer');
    $view->append($processed_shelves_template);

    ##################



    # 1 Retrieve book list
    # 2 Send each book URL to the Curl Requester
    # 3 Each flux is stored in a file
    my $list_vars;

    foreach my $list_id( @OL_list_ids ){
        # DEBUG
        # print $OL_list_idx . "\n";


        if ( $microlibrary::html_mode ){



        # !!! List = shelves
        $list_vars = {
            # Library infos
            'library' => {
                # Made to display the library name into the shelves.tmpl template
                'name'              =>  $microlibrary::context{lib_display_name},
            },
            'shelf'   => {
                'name'              =>  "[% shelf.name %]",
                'description'       =>  $microlibrary::context{shelf}{description},
                'id'                =>  $list_id,
                # Another hash into the hash could be stored ???
                # ...  
            },
        };

        # say $$list_vars{shelf}{name};

        if ( $microlibrary::log_mode ){
            say Dumper $list_vars;
        }

        ###############
        #
        #       BOOK SHELF OPENING  
            
        $view->just_fucking_print("<section class=\"shelf\" id=\"$$list_vars{shelf}{id}\">");
        
        ## BUG !!!!!!!! Can't get the LIST TITTLE !!!
        $view->just_fucking_print("<h2 class=\"shelf-label\">[% shelfsize.$OL_list_idx.name %]</h2>"); #  . $$list_vars{shelf}{name} . 

        }

        # Generate tmp buffer for the books and shelfes templates
        $rdf_file_containing_books_list->parse_RDF_file(
            $OL_list_ids[ $OL_list_idx ]
        );
        if ( $microlibrary::html_mode ){
            
            #$view->close_tag();

        $view->just_fucking_print("<!-- Shelf closing --></section>");

        }

        $OL_list_idx ++;

    }

     
    my $size = @shelves;
    
    # say Dumper @shelves;

    if( $microlibrary::log_mode){
        foreach my $title (@shelves){
            if (defined $title){
                say "Title = " . $title;
            }        
        }
    }

    if ( $microlibrary::html_mode ){

                ###############
        #
        #       BOOK SHELF CLOSING   


        #$view->close_tag( "newline" );
        #$view->print_into_tags("back to top of the world", "a", "href", "#head", "class", "anchor-link");
        #$view->close_tag( "newline" );

        &print_generic_footer();
        # Closing .?
        $view->close_tag( "newline", "end", ".?" );
        # Closing #page
        $view->close_tag( "newline", "end", "#page" );
        $view->close_tag( "newline", "end", "site" );
    }

}

# This is small contact, fuck or about pages, no logic
else {
    my $template_vars = {
        'menu' => {
            # Made to display the library name into the shelves.tmpl template
            'nav' => {
                $microlibrary::context{home_page_flag} =>  $microlibrary::context{home_page_flag},
            },             
        },
    };


    if ( $context{'home_page_flag'} eq "about" || $context{'home_page_flag'} eq "contact" || $context{'home_page_flag'} eq "fuck"  ){

        my $template = Template->new($microlibrary::templates_conf);

        my $template_file = $microlibrary::context{home_page_flag}.".tmpl";
        my $tmp_template_buffer = '';

        $template->process($template_file, $template_vars, 'tmp_template_buffer')
            || die $template->error("Can't process template :(");

        # say "Template processed";

        my $processed_template = $view->append_tmp_tmpl();
        $view->append($processed_template);

        &print_generic_footer;

    }
}



(my $end_epoch_seconds, my $end_epoch_microseconds) = gettimeofday;
(my $end_second, my $end_minute, my $end_hour) = localtime($end_epoch_seconds);
my $display_ending_time =  sprintf("%02d:%02d:%02d", $end_hour, $end_minute, $end_second) . "\n"; 

my $total_seconds_time = $end_epoch_seconds - $begin_epoch_seconds;
my $total_microseconds_time = ( $end_epoch_microseconds - $begin_epoch_microseconds ) / 1_000_000;


# If needed : printf "%f", 13.534e26 to convert an exponetial number to decimal number
my $total_time = $total_seconds_time + $total_microseconds_time;





if ( $microlibrary::log_mode ){
    print "\n\nSTATS :\n   Beginning process : $display_beginning_time\n   Ending process : $display_ending_time\n";
    print "   Script executed in $total_time s.";
}

if( !$query_string_param_ref ){
    my $query_string_param = "UNDEF !"
}


# Là c'est n'importe quoi, il y a des trucs qui ne devraient pas du tout être là...
if ( $microlibrary::browser_debug_mode && $microlibrary::html_mode  ){

    my $stats = "<div class=\"stats\">
                Beginning process : 
                <span>$display_beginning_time</span><br />
                Ending process : 
                <span>$display_ending_time</span><br />
                <span>home_page_flag = $home_page_flag</span><br />
                <span>library = $nick</span><br />
                <span>query_string = TO DO</span><br />
                <span>Books nbr. for last $nick\'s list = $microlibrary::context{items_nbr}</span><br />
                <span>Page generated in $total_time s.</span><br />
                <span><a href=\"http://localhost:81/~foutaiz/lib/cookies.pl\">The cookie buster</a></span>
            </div>";

     if ( $microlibrary::log_mode ){
         print Dumper $stats;
     }

     $microlibrary::html_buffer->append( $stats );
     if ( $microlibrary::log_mode ){
         print Dumper $html_buffer;
     }
}


if ( $microlibrary::html_mode  ){
    # Put JS libraries
    if ( $microlibrary::html_mode ){
        # Should...
        $view->insert_JS();
        # ... but I fuck it !
        #my $scripts = '<script type="text/javascript" src="./View/js/jquery-1.4.3.min.js"></script><script type="text/javascript" src="./View/js/scroller.js"></script>';
    }


    # Body closing tag
    if ( $microlibrary::html_mode ){
        $view->close_tag( "newline", "end", "?" );
        # Body closing
        $view->close_tag( "newline", "end", "body" );
    }

    # Just print </html>
    #$view->close_footer();
    $view->include_tmpl("doctype_closing");


    # ######################################################################"
    #
    # Attention aux droits sur ce fichier !!!!! 
    # Il est possible qu'ils soit nécessaire qu'ils soient RW-R--RW ou 666
    # Une série de tests complets devra être mise en place à ce sujet :
    #   -> le fichier n'existe pas
    #   -> les droits du fichier ne sont pas les bons
    #   -> etc.
    #
    my $cache_update = 0;
    if ( $html_buffer->save("../cache/html_buffer.txt")){
        $cache_update = 1;
        if ( $microlibrary::log_mode ){
            print "Cache file has been updated !!!";
        }
 
    } else {
        $cache_update = 0;
        if ( $microlibrary::browser_debug_mode ){
           if ($html_buffer->isError()) { 
              my $error = "Error: " . $html_buffer->getError() . "\n"; 
              print "Error: " . $html_buffer->getError() . "\n";
              print "Impossible to write into the cache file. Please fix write rights! 666 is a good solution..."
           } 
        }
    }

    my $html_final_buffer = FileHandle->new;

    #######
    #
    #  V0.02.1
    #  We generate the HTML buffer with templating stuff inside
    #  Before being send to the browser, the HTML buffer has to be
    #  rendered, in the Template Toolkit way, cause we print some 
    #  [% INCLUDES %] inside.
    #
    #######


    if ($html_final_buffer->open("< ../cache/html_buffer.txt")) {
        ###########################################################################
        #
        #       TEMPLATE TOOLKIT RENDERING HERE !!!!
        #                   T O D O
        #
        #           print $processed_templates;
        
        my $templates = Template->new($templates_conf);

        # Add vars like menu stuff or debug things here !!!
        # ... to do so, just add a $var argument to the process()
        my $final_template_vars = {
            # menu infos
            'menu' => {
                # Made to display the library name into the shelves.tmpl template
                'nav' => {
                    $microlibrary::context{home_page_flag} =>  $microlibrary::context{home_page_flag},
                },             
            },
        };

        my $tmp_size = 0;
        my $tmp_tmp_size = 0;
        my $shelves_number = @shelves; 
        my $concatenate_size;


        # my $shelf_size_name = {
        #     'shelfsize' => {
        #         $tmp_size => {
        #             'name' => $shelves[1],
        #         },
        #     },
        # };

        # say Dumper $shelf_size_name;

        # $tmp_size++;

        while ($tmp_size < $shelves_number ){
            
            foreach my $title (@shelves){ 
            
                if (defined $title){
                    # say $title . "<br />";

                    $$final_template_vars{'shelfsize'}{$tmp_tmp_size}{'name'} = $title;

                    $tmp_tmp_size++;

                }
                $tmp_size++;

            }

            # $final_template_vars = { 
            #     $shelf_size_name };

        }

        # say Dumper $final_template_vars;

        # say  $$final_template_vars{shelfsize}{0}{name};
        # say  $$final_template_vars{shelfsize}{1}{name};
        # say  $$final_template_vars{shelfsize}{2}{name};

        if ( $microlibrary::log_mode ){
            say Dumper $microlibrary::context{home_page_flag};
            say Dumper $final_template_vars;
        }


        $templates->process($html_final_buffer, $final_template_vars)
            || die $templates->error("Can't process templates in the final processing :(");
        #
        ###########################################################################
        print <$html_final_buffer>; # Generates XHTML output
        if( $cache_update == 1 ){
            if ( $microlibrary::browser_debug_mode ){
                print "<span class=\"updated\">:) Cache file HAS BEEN updated</span>";
            }
        } elsif( $cache_update == 0 ) {
            if ( $microlibrary::browser_debug_mode ){
                my $error="";
                if ($html_buffer->isError()) { 
                    $error = "Error: " . $html_buffer->getError() . "\n"; 
                    print "Error: " . $html_buffer->getError() . "\n";
                    print "Impossible to write into the cache file. Please fix write rights! 666 is a good solution..."
                } 
                print "<span class=\"non-updated\"><blink>:( Cache file has NOT been updated\n $error</blink></span>";
            }
        }
        $html_final_buffer->close;
    } else {
        die("Oups, the file containing cached html don't exists...");
    }
}

if ( $microlibrary::sad_deps_mode ){
    my $env_viewer = Tools::DependanciesViewer->new();
    $env_viewer->print_used_modules($0, %INC);
    # Could be a more elegant way to do it...
    #my $hash_ref_scan = scan_deps(
        #files => [ 'microlibrary.pl' ],
        #recurse => 1,
    #);
}


# Should be located into View::WebDisplayer, not here, but I've been too lazy
# to put it in the right place :(
sub print_generic_footer{

    # Should process a short browser name
    my $his_browser = $ENV{ 'HTTP_USER_AGENT' };
    my $his_IP = $ENV{ 'REMOTE_ADDR' };
    
    #######################
    #
    #       TEMPLATE
    #
    # ALWAYS USE THE VARS, DON'T TRY TO PROCESS WITHOUT VARS
    my $footer_vars = {
        site        =>{
            host            => $host,
            browser         => $his_browser,
            IP              => $his_IP,
        },
    };

    my $footer_template = Template->new($microlibrary::templates_conf);

    my $footer_template_file = 'footer.tmpl';
    # say "ok";
    # my $tmp_template_buffer = '';

    # Add ""
    $footer_template->process($footer_template_file, $footer_vars, 'tmp_template_buffer')
        || die $footer_template->error();

    #say "ok";

    # say "Template processed";

    my $processed_template = $view->append_tmp_tmpl();
    $view->append($processed_template);

    
    $microlibrary::html_buffer->append( "&nbsp;" );

}
#########  ????????????????
}
__END__

=pod

=head1 NAME The Microlibrary mimicrokernel


=head1 DESCRIPTION

This script runs a web-service thru the Open Library (OL) API, stores data send by
the web-service into files and generate XHTML to display data in a very simple page.
The microlibrary is made to manage everything to generate a book-shelf page, using 
your data from Open Library. For more informations and How To, please see http://goo.gl/k375R

=head2 The Self Shelf Swap Project

This first implementation is a physical object (books) swaping/sharing system.
Our Library project is to create an online index which could make (in next version) possible to : 

=over 4

=item Inform

retrieve informations about registered documents 

=item Locate

know where documents are physically based and if you can loan it )

=item Create

register an account & create your own Library DataBase

=item Share

share it with other users.

=back

Our desire is to provide a private Libraries Network in an IRL spirit, a Library 
of Libraries ("meta-libraries"), physically based (shelfs). 
We don't want to provide downloadable documents.

That's why our project is named "Self Shelf Swap". It's first implementation will be Bibliothek.

=head2 How does-it do the job ?

=over 4

=item B<< The lists >>

List are considered like thematic shelfes of your library.
First the microlibrary mimicrokernel calls the Properties Reader, that will parse our open_library_lists file, 
and return an array if list ids. This array is passed as an argument to the 
build_the_list_request() subroutine of the Curl Requester, that will build 
B<< and >> get the list requests. For each list id, we get a RDF flux containing 
books URI. Flux are stored into your hard drive by passing the RDF flux, the 
list id and the "list" flag (to specify that we manage lists) to the store_RDF_flux() 
of the RDF Tool. 

=item B<< The books >>

Next, for each list, we load the file on disk thru the parse_RDF_file() subroutine, 
passing the list id as an argument. We get many data thru the get_list_element()
subroutine of RDFTool, that will get all books data. For more information about 
this part, see the RDFTool documentation.

=item B<< The View >>

The Web Displayer was and is still scattered everywhere. And will be remplaced soon by a more decent
system (AKA MVC). Now, at 0.02 version it's a little bit better cause we put any view item in a buffer
for later usage. See you soon...

=back

=head2 HOW TO SET UP YOUR MICROLIBRARY

=head3 INTERACT WITH OPEN LIBRARY AND GENERATE YOUR DATABASE

To deploy and install your microlibrary, you need a GNU/Linux environment or any
Perl friendly system. Dime and a shoestring has been only tested on Debian or Ubuntu.
Feel free to test it on other platforms and give us feedback...

Install every dependencies. You can follow the DEPENDENCIES section of this documentation
and test it by running the microlibrary.pl script. To run this script, just open 
a terminal, go to the dime_and_a_shoestring. Fix the $ENV{HOME} environment variable
that should target the 'lib' directory containing the microlibrary.pl script.
Set the $reload_mode and the $log_mode flags to '1', let the $html_mode and $sad_deps_mode
to '0'.

Now type into your terminal :

    $ perl microlibrary.pl
    
The script should start and send querries to microlibrary. This can take some time.
We wait to receive the response of every querry before to send another one. Once 
every  querries have been send, a kind of database is stored into your hard drive in

    /home/foo/bar/dime_and_a_shoestring/database/book/$ 
                                                /list/$
=head3 MAKE A VERY SIMPLE WEB SITE

Now turn the $reload_mode and the $log_mode flags to '0'. As your database is on your own disk now
you don't even need to send request to Open Library again. You will only re-run the script with $reload_mode 
activated each time you add a list/shelf, a book, or update informations into OL.
Set the $html_mode to '1'. Your microlibrary is ready, to be used from a browser you just need to deploy it 
on your public server (don't forget to update the $ENV{HOME} variable).

=head1 CONFIGURATION AND ENVIRONMENT

=head2 Flags for modes

All the flags for modes should be configurated from the /properties/modes.properties file.

=over 4

=item B<< Reload OL data mode >> 

We will use the Open Library database and list feature to constitute the shelfes
of a very simple library : lists are thematic shelfes, and your set of lists under
a OL account is a Microlibrary. Soon, we will provide more powerful features.
You can turn of the interraction with the OL API by giving the value of 0 to the 
$reload_mode scalar. Data will be read from the files stores on your harddrive.
Script with "$reload_mode = 1" must have been run at least once to generate files 
on your hard drive.

=item B<< Display log mode >>

In development and debug environment you must most of the time use our "log mode" 
that will display useful informations when processing. Don't forget to turn it 
of ( $log_mode = 0) in production environment cause it will display logs into you web pages.
We will soon replace this by a more descent and classical LogForPerl logger.

=item B<< Browser simulation mode >>

Maybe it is bad-named. Simulate a browser querry on the main script. It generates a query string ("?lib=lib_name") when you're 
working from console. It should be used to reload the datas from open-library for a particular
user. You should set the  $nick scalar to select wich library will be reload. Experience makes
us learn that you should avoid to use this mode in terminal work, near of deploiement time, 
and prefer a clean and reasonable debug direclty from your browser. That's why the  
'Browser debug mode has been made' !

=item B<< Display XHTML tags mode >>

This is the mode for production environment. It generate your data into XHTML
 tags ( $html_mode = 1 ). Data are stored into a temporary buffer, that will be soon
print when asked by a browser. It can also be used into console for debug, combinated with
the browser simulation mode.

=item B<< SAD dependencies mode >>

This is the Simple And Dirty dependencies logger ( $sad_deps_mode = 1 ), that 
will display all needed dependencies. Only works when dependencies are used (to confirm ?)
Don't forget to turn it of ( $sad_deps_mode = 0 ) in production environment cause 
it will display logs into you web pages :(.

=item B<< Browser debug mode >>

Will simply display useful debug things into the browser into an independant fixed <div /> that 
will not smashed the delicate display of your pages and gives you performances stuffs and 
other techs goodies. MUST be turned off in production environment. It has been proved that's
it's better to test as soon as possible into a browser. In early development time you
can use the browser debug mode but it's  more sure to switch to a real browser as soon as possible.
Browser debug mode is good for terminal kiddies.

=back

=head1 VERSION

This is version 0.02.1, well known as a the Chinese Loony And A Rope

=head1 DEPENDENCIES

=head2 Self Shelf Swap dependencies

Microlinrary comes with :
Data::CURL::CurlRequester - Data:Properties::PropertiesReader - Data::RDF::RDFTool
Tools::DependanciesViewer - View::WebDisplayer 

=head2 CPAN dependencies

We use these CPAN modules : 
Data::Dumper - Encode::Encoder - File::Basename - IO::File 
- Module::ScanDeps - Net::Curl::Simple - Properties - XML::Simple
Time::HiRes

This will at least need these following CPAN modules to work : 
local::lib - Log::Simplest - Merror - RDF::Redland - Test::Simple


=head1 SUBROUTINES / METHODS

=over 4

=item B<< print_generic_footer() >>

It seems to be self explanatory, but what is not is it should be located 
into View::WebDisplayer, not here, but I've been too lazy to move it :(, it will be made into our 
next *smart* version, soon available in 2027.

=back

=head1 ARGUMENTS

I would like it could take parameters in command line mode, just like :

    $ perl microlibrary.pl --reload

... to force the reload mode to be activated for ALL lists and books

    $ perl microlibary.pl --reload --name foo

... will reload only the lists and books of a user. 'foo' will be supposed to be 
a valid Open Library user name, listed into our open_library_lists.properties file.

Even better, if this user don't exist, it could be created :

   $ User 'foo' don't exist, should I create it ? (yes / no)

... and add it into the properties file. Will load all the user's lists, put it the
properties files and make all the stupid work that you don't want to do. Out in 2054 !!!


=head1 BUGS AND LIMITATIONS

The Microlibrary mimicrokernel 0.01 version has been an exercice for learning perl and is not 
recommended for re-use. Even if we try to make it better for the 0.02 version (it's really better),
lots of work should be done to reach the perfect 0.03 version and the wonderful mythical 1.0 version.

=head2 Hard coded things and configuration stuff

Some hard coded things and configuration stuff should be into properties files.
See # See https://bugs.launchpad.net/bibliothek-project/+bug/856571

=head2 Non-portability

The Microlibrary is not cross-platform and has been tested ONLY on Debian-like 
GNU/Linux systems. At this day, we didn't try to deploy it on MS Windows, but feel
free to help us if you want !

=head2 Scattering view !!!

This is the most S.A.D. thing about the microlibrary... View is scattered everywhere,
there is nothing to do than do re-do it correctly !!!

=head2 Dependancies section

This section could be ported to a meta.yaml file


=head1 AUTHOR 

Self Shelf Swap - Smonff - 2011-06-29

=head1 CURRENT MAINTAINER

    my ( $nickname = "smonff", $provider = "lavabit.com");
    print $nickname . "@" . $provider;

=head1 COPYRIGHT AND LICENSE

Copyright 2012 Self Shelf Swap.

This library is free software. You can redistribute it and/or modify
it under the same terms as Perl himself.

=cut
